# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
from .callbacks import Callback, Callbacks, Cbs
import k1lib, time
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["BatchLimit", "EpochLimit", "TimeLimit", "CancelOnExplosion",
           "CancelOnLowLoss", "CancelOnHighAccuracy", "CancelOnOverfit", "DontTrain",
           "GradientClipping", "GradientClippingNorm", "TrainOnly", "ValidOnly"]
@k1lib.patch(Cbs)
class BatchLimit(Callback):                                                      # BatchLimit
    """Cancels the epoch after executed certain number of batches"""             # BatchLimit
    def __init__(self, limit:int):                                               # BatchLimit
        super().__init__(); self.order = 25                                      # BatchLimit
        self.limit = limit if limit != None else float("inf")                    # BatchLimit
    def startEpoch(self): self.currentBatch = 0                                  # BatchLimit
    def startBatch(self):                                                        # BatchLimit
        if self.currentBatch >= self.limit:                                      # BatchLimit
            raise k1lib.CancelEpochException(f"Batch {self.limit} reached")      # BatchLimit
    def endBatch(self): self.currentBatch += 1                                   # BatchLimit
@k1lib.patch(Cbs)                                                                # BatchLimit
class EpochLimit(Callback):                                                      # EpochLimit
    """Cancels the run after executed certain number of epochs"""                # EpochLimit
    def __init__(self, limit:int):                                               # EpochLimit
        super().__init__(); self.order = 25                                      # EpochLimit
        self.limit = limit if limit != None else float("inf")                    # EpochLimit
    def startRun(self): self.currentEpoch = 0                                    # EpochLimit
    def startEpoch(self):                                                        # EpochLimit
        if self.currentEpoch >= self.limit:                                      # EpochLimit
            raise k1lib.CancelRunException(f"Epoch {self.limit} reached!")       # EpochLimit
    def endEpoch(self): self.currentEpoch += 1                                   # EpochLimit
@k1lib.patch(Cbs)                                                                # EpochLimit
class TimeLimit(Callback):                                                       # TimeLimit
    """Cancels the run after a certain number of seconds have passed"""          # TimeLimit
    def __init__(self, seconds=30):                                              # TimeLimit
        super().__init__(); self.seconds = seconds if seconds != None else float("inf"); self.order = 25 # TimeLimit
    def startRun(self): self.startTime = time.time()                             # TimeLimit
    def startBatch(self):                                                        # TimeLimit
        if time.time() - self.startTime > self.seconds:                          # TimeLimit
            raise k1lib.CancelRunException(f"Takes more than {self.seconds} seconds!") # TimeLimit
@k1lib.patch(Cbs)                                                                # TimeLimit
class CancelOnExplosion(Callback):                                               # CancelOnExplosion
    """Cancels the run if any of the parameters are larger than a certain limit""" # CancelOnExplosion
    def __init__(self, limit:float=1e6):                                         # CancelOnExplosion
        super().__init__(); self.order = 25                                      # CancelOnExplosion
        self.limit = limit; self.triggered = False                               # CancelOnExplosion
    def startRun(self): self.triggered = False                                   # CancelOnExplosion
    def startBatch(self):                                                        # CancelOnExplosion
        for p in self.l.model.parameters():                                      # CancelOnExplosion
            o = p.detach()                                                       # CancelOnExplosion
            if o.max().float() > self.limit or o.min().float() < -self.limit:    # CancelOnExplosion
                self.triggered = True                                            # CancelOnExplosion
                raise k1lib.CancelRunException("Explosion detected!")            # CancelOnExplosion
    def __repr__(self):                                                          # CancelOnExplosion
        return f"""{self._reprHead}, use...
- cb.triggered: to see if there was an explosion on the last run
- cb.progress: to see current progress at explosion time
{self._reprCan}"""                                                               # CancelOnExplosion
@k1lib.patch(Cbs)                                                                # CancelOnExplosion
class CancelOnLowLoss(Callback):                                                 # CancelOnLowLoss
    " "                                                                          # CancelOnLowLoss
    def __init__(self, loss:float, epochMode:bool=False):                        # CancelOnLowLoss
        """Cancels the run if loss is lower than amount specified.
Original class: :class:`~k1lib.callbacks.limits.CancelOnLowLoss`

:param epochMode: False if use batch loss, True if use valid epoch loss"""       # CancelOnLowLoss
        super().__init__(); self.order = 25; self.dependsOn = ["Loss"]           # CancelOnLowLoss
        self.loss = loss; self.epochMode = epochMode                             # CancelOnLowLoss
    def startRun(self):                                                          # CancelOnLowLoss
        if not hasattr(self.l.cbs, "Loss"):                                      # CancelOnLowLoss
            raise AttributeError("Learner does not have required `Loss` callback") # CancelOnLowLoss
        self.v = self.cbs.Loss.valid; self.ve = self.cbs.Loss.epoch.valid # List[int] # CancelOnLowLoss
    def endBatch(self):                                                          # CancelOnLowLoss
        if self.epochMode:                                                       # CancelOnLowLoss
            if len(self.ve) > 0 and self.ve[-1] < self.loss:                     # CancelOnLowLoss
                raise k1lib.CancelRunException(f"Low loss {self.loss} ({self.ve[-3:]} actual) achieved!") # CancelOnLowLoss
        elif len(self.v) and self.v[-1] < self.loss:                             # CancelOnLowLoss
            raise k1lib.CancelRunException(f"Low loss {self.loss} ({self.v[-3:]} actual) achieved!") # CancelOnLowLoss
@k1lib.patch(Cbs)                                                                # CancelOnLowLoss
class CancelOnHighAccuracy(Callback):                                            # CancelOnHighAccuracy
    """Cancels the run if accuracy is higher than the amount specified"""        # CancelOnHighAccuracy
    def __init__(self, accuracy:float):                                          # CancelOnHighAccuracy
        super().__init__(); self.order = 25                                      # CancelOnHighAccuracy
        self.accuracy = accuracy; self.dependsOn = ["Accuracy"]                  # CancelOnHighAccuracy
    def endBatch(self):                                                          # CancelOnHighAccuracy
        if not hasattr(self.l, "Accuracy"): raise AttributeError("Learner does not have `Accuracy` callback") # CancelOnHighAccuracy
        a = self.l.Accuracy.valid[-1]                                            # CancelOnHighAccuracy
        if a > self.accuracy:                                                    # CancelOnHighAccuracy
            raise k1lib.CancelRunException(f"High accuracy {self.accuracy} ({a} actual) achieved!") # CancelOnHighAccuracy
@k1lib.patch(Cbs)                                                                # CancelOnHighAccuracy
class CancelOnOverfit(Callback):                                                 # CancelOnOverfit
    def __init__(self, ratio:float=1.2, alpha:float=0.99, after:int=10):         # CancelOnOverfit
        """Cancels the run if overfit is detected.

:param ratio: Max ratio between the lowest loss and the current loss before cancelling the run
:param alpha: Moving average's alpha, used for both minLoss and loss estimates
:param after: After how many epochs should the overfit detection be activated?""" # CancelOnOverfit
        super().__init__(); self.ratio = ratio                                   # CancelOnOverfit
        self.minLoss = k1lib.MovingAvg(alpha=alpha, debias=True)                 # CancelOnOverfit
        self.loss = k1lib.MovingAvg(alpha=alpha, debias=True)                    # CancelOnOverfit
        self.count = 0; self.after = after                                       # CancelOnOverfit
    def startRun(self): self.count = 0                                           # CancelOnOverfit
    def endEpoch(self): self.count += 1                                          # CancelOnOverfit
    def endBatch(self):                                                          # CancelOnOverfit
        if not self.l.model.training:                                            # CancelOnOverfit
            loss = self.l.loss; self.loss(loss)                                  # CancelOnOverfit
            if self.loss.value < self.minLoss.value or self.minLoss.value == 0: self.minLoss(self.loss.value) # CancelOnOverfit
            if self.count > self.after and self.loss.value > self.minLoss.value * self.ratio: # CancelOnOverfit
                raise k1lib.CancelRunException(f"Overfit detected! Smoothed min loss: {self.minLoss.value}, loss: {loss}") # CancelOnOverfit
@k1lib.patch(Cbs)                                                                # CancelOnOverfit
class DontTrain(Callback):                                                       # DontTrain
    """Don't allow the network to train at all"""                                # DontTrain
    def startBackward(self): return True                                         # DontTrain
    def startStep(self): return True                                             # DontTrain
if hasTorch:                                                                     # DontTrain
    from torch.nn.utils import clip_grad_value_                                  # DontTrain
    @k1lib.patch(Cbs)                                                            # DontTrain
    class GradientClipping(Callback):                                            # DontTrain
        """Clips gradient to a specific max value"""                             # DontTrain
        def __init__(self, value:float): super().__init__(); self.value = value  # DontTrain
        def startStep(self):                                                     # DontTrain
            clip_grad_value_(self.l.model.parameters(), self.value)              # DontTrain
else:                                                                            # DontTrain
    class GradientClipping(Callback): pass                                       # DontTrain
if hasTorch:                                                                     # DontTrain
    from torch.nn.utils import clip_grad_norm_                                   # DontTrain
    @k1lib.patch(Cbs)                                                            # DontTrain
    class GradientClippingNorm(Callback):                                        # DontTrain
        """Clips gradient to a specific max_norm value. Can choose to lump
all params together or do each separately.

See also: :class:`~k1lib.callbacks.limits.GradientClipping` callback."""         # DontTrain
        def __init__(self, max_norm:float, each:bool=True):                      # DontTrain
            super().__init__(); self.max_norm = max_norm; self.each = each       # DontTrain
        def startStep(self):                                                     # DontTrain
            if self.each:                                                        # DontTrain
                for m in self.l.model.parameters():                              # DontTrain
                    clip_grad_norm_(m, self.max_norm)                            # DontTrain
            else: clip_grad_norm_(self.l.model.parameters(), self.max_norm)      # DontTrain
else:                                                                            # DontTrain
    class GradientClippingNorm(Callback): pass                                   # DontTrain
@k1lib.patch(Cbs)                                                                # DontTrain
class TrainOnly(Callback):                                                       # TrainOnly
    " "                                                                          # TrainOnly
    def __init__(self, cb):                                                      # TrainOnly
        """Only executes specified callback when training. This modifies the callback's
``suspended`` variable, so it may interfere with :meth:`k1lib.callbacks.callbacks.Callbacks.suspend`
by setting it to different values while in the context."""                       # TrainOnly
        super().__init__(); self.cb = cb                                         # TrainOnly
    def startBatch(self):                                                        # TrainOnly
        self.cb.suspended = not self.l.model.training                            # TrainOnly
@k1lib.patch(Cbs)                                                                # TrainOnly
class ValidOnly(Callback):                                                       # ValidOnly
    " "                                                                          # ValidOnly
    def __init__(self, cb):                                                      # ValidOnly
        """Same as :class:`TrainOnly`, but only executes specified callback when doing
validation."""                                                                   # ValidOnly
        super().__init__(); self.cb = cb                                         # ValidOnly
    def startBatch(self):                                                        # ValidOnly
        self.cb.suspended = self.l.model.training                                # ValidOnly



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This is for functions that are actually biology-related
"""
from k1lib.cli.init import BaseCli
import k1lib; import k1lib.cli as cli
import os; from functools import partial
from typing import Iterator, Union
__all__ = ["go", "quality", "longFa", "idx",
           "transcribe", "complement", "translate", "medAa", "longAa"]
settings = k1lib.Settings()
k1lib.settings.cli.add("bio", settings, "from k1lib.cli.bio module");
def _patchDir(term, s, p):                                                       # _patchDir
    if p != None: p = os.path.abspath(os.path.expanduser(p))                     # _patchDir
    s.__dict__[term] = p                                                         # _patchDir
settings.add("blast", None, "location of BLAST database", partial(_patchDir, "blast")) # _patchDir
settings.add("go", None, "location of gene ontology file (.obo)", partial(_patchDir, "go")) # _patchDir
settings.add("so", None, "location of sequence ontology file", partial(_patchDir, "so")); # _patchDir
settings.add("lookupImgs", True, "sort of niche. Whether to auto looks up extra gene ontology relationship images") # _patchDir
def go(term:int):                                                                # go
    """Looks up a GO term"""                                                     # go
    if settings.go is None and not os.path.exists("go.obo"):                     # go
        answer = input("""No gene ontology obo file specified! You can:
- Specify the file using `settings.cli.bio.go = '/some/folder/go.obo'`
- Download this automatically to file `go.obo`

You want to download this automatically? (y/n) """)                              # go
        if answer.lower().startswith("y"):                                       # go
            url = "http://current.geneontology.org/ontology/go.obo"              # go
            print(f"Downloading from {url}...      ", end="")                    # go
            cli.wget(url); print("Finished!")                                    # go
        else: return print("Aborted")                                            # go
    file = settings.go or "go.obo"; term = f"{term}".rjust(7, "0")               # go
    cli.cat(file) | cli.grep(f"id: GO:{term}", 0, 10) > cli.stdout()             # go
    print(f"https://www.ebi.ac.uk/QuickGO/GTerm?id=GO:{term}")                   # go
    if settings.lookupImgs:                                                      # go
        class Repr:                                                              # go
            def _repr_html_(self):                                               # go
                return f"""<img src="http://amigo.geneontology.org/visualize?mode=amigo&term_data_type=string&format=png&inline=false&term_data=GO%3A{term}" />""" # go
        return Repr()                                                            # go
settings.add("phred", """!"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJ""", "Phred quality score") # go
class toIdx(BaseCli):                                                            # toIdx
    def __init__(self, chars:str): self.chars = {v:k for k, v in enumerate(chars)} # toIdx
    def __ror__(self, it):                                                       # toIdx
        chars = self.chars                                                       # toIdx
        for e in it: yield chars[e]                                              # toIdx
def quality(log=True):                                                           # quality
    """Get numeric quality of sequence.
Example::

    # returns [2, 2, 5, 30]
    "##&?" | quality() | deref()

:param log: whether to use log scale (0 -> 40), or linear scale (1 -> 0.0001)""" # quality
    if log: return toIdx(settings.phred)                                         # quality
    else: return toIdx(settings.phred) | cli.apply(lambda x: 10**(-x/10))        # quality
def longFa():                                                                    # longFa
    """Takes in a fasta file and put each sequence on 1 line.
File "gene.fa"::

    >AF086833.2 Ebola virus - Mayinga, Zaire, 1976, complete genome
    CGGACACACAAAAAGAAAGAAGAATTTTTAGGATC
    TTTTGTGTGCGAATAACTATGAGGAAGATTAATAA
    >something other gene
    CGGACACACAAAAAGAAAGAAGA
    TTTTGTGTGCGAATAACTATGAG

Code::

    cat("gene.fa") | bio.longFa() | cli.headOut()

Prints out::

    >AF086833.2 Ebola virus - Mayinga, Zaire, 1976, complete genome
    CGGACACACAAAAAGAAAGAAGAATTTTTAGGATCTTTTGTGTGCGAATAACTATGAGGAAGATTAATAA
    >something other gene
    CGGACACACAAAAAGAAAGAAGATTTTGTGTGCGAATAACTATGAG"""                            # longFa
    return cli.grep("^>", sep=True).till() | (cli.item() & (~cli.head(1) | cli.join(""))).all() | cli.joinStreams() # longFa
def _fileWithoutExt(f): return ".".join(f.split(".")[:-1])                       # _fileWithoutExt
class idx(BaseCli):                                                              # idx
    """Indexes files with various formats."""                                    # idx
    @staticmethod                                                                # idx
    def blast(fileName:str=None, dbtype:str=None):                               # idx
        """Uses ``makeblastdb`` to create a blast database from a fasta file.
Example::

    "file.fa" | bio.idx.blast()
    bio.idx.blast("file.fa")"""                                                  # idx
        f = cli.applyS(lambda fileName: None | cli.cmd(f"makeblastdb -dbtype {dbtype or 'nucl'} -in {fileName} -out {_fileWithoutExt(fileName)}")) # idx
        return f if fileName is None else f(fileName)                            # idx
    @staticmethod                                                                # idx
    def bwa(fileName:str=None):                                                  # idx
        """Uses ``bwa`` to index a fasta file.
Example::

    "file.fa" | bio.idx.bwa()
    bio.idx.bwa("file.bwa")"""                                                   # idx
        f = cli.applyS(lambda fileName: None | cli.cmd(f"bwa index {fileName}")) # idx
        return f if fileName is None else f(fileName)                            # idx
    @staticmethod                                                                # idx
    def bam(fileName:str=None):                                                  # idx
        """Uses ``samtools`` to index a bam file.
Example::

    "file.bam" | bio.idx.bam()
    bio.idx.bam("file.bam")"""                                                   # idx
        f = cli.applyS(lambda fileName: None | cli.cmd(f"samtools index {fileName}")) # idx
        return f if fileName is None else f(fileName)                            # idx
class transcribe(BaseCli):                                                       # transcribe
    """Transcribes (DNA -> RNA) incoming rows.
Example::

    # returns "AUCG"
    "ATCG" | transcribe()
    # returns ["AUCG"]
    ["ATCG"] | transcribe() | deref()"""                                         # transcribe
    def __ror__(self, it:Union[Iterator[str], str]):                             # transcribe
        if isinstance(it, str): return [it] | self | cli.item()                  # transcribe
        return (line.upper().replace("T", "U") for line in it)                   # transcribe
class complement(BaseCli):                                                       # complement
    """Get the reverse complement of DNA.
Example::

    # returns "TAGC"
    "ATCG" | bio.complement()
    # returns ["TAGC"]
    ["ATCG"] | bio.complement() | deref()"""                                     # complement
    def __ror__(self, it:Union[Iterator[str], str]):                             # complement
        if isinstance(it, str): return [it] | self | cli.item()                  # complement
        return (line.upper().replace("A", "0").replace("T", "A").replace("0", "T").upper().replace("C", "0").replace("G", "C").replace("0", "G") for line in it) # complement
ntAa = {"UUU": "F", "UUC": "F", "UUA": "L", "UUG": "L",                          # complement
        "UCU": "S", "UCC": "S", "UCA": "S", "UCG": "S",                          # complement
        "UAU": "Y", "UAC": "Y", "UAA": "*", "UAG": "*",                          # complement
        "UGU": "C", "UGC": "C", "UGA": "*", "UGG": "W",                          # complement
                                                                                 # complement
        "CUU": "L", "CUC": "L", "CUA": "L", "CUG": "L",                          # complement
        "CCU": "P", "CCC": "P", "CCA": "P", "CCG": "P",                          # complement
        "CAU": "H", "CAC": "H", "CAA": "Q", "CAG": "Q",                          # complement
        "CGU": "R", "CGC": "R", "CGA": "R", "CGG": "R",                          # complement
                                                                                 # complement
        "AUU": "I", "AUC": "I", "AUA": "I", "AUG": "M",                          # complement
        "ACU": "T", "ACC": "T", "ACA": "T", "ACG": "T",                          # complement
        "AAU": "N", "AAC": "N", "AAA": "K", "AAG": "K",                          # complement
        "AGU": "S", "AGC": "S", "AGA": "R", "AGG": "R",                          # complement
                                                                                 # complement
        "GUU": "V", "GUC": "V", "GUA": "V", "GUG": "V",                          # complement
        "GCU": "A", "GCC": "A", "GCA": "A", "GCG": "A",                          # complement
        "GAU": "D", "GAC": "D", "GAA": "E", "GAG": "E",                          # complement
        "GGU": "G", "GGC": "G", "GGA": "G", "GGG": "G"}                          # complement
_shortAa = {v:v for v in ntAa.values()}                                          # complement
_medAa = {                                                                       # complement
    "F": "Phe", "L": "Leu", "I": "Ile", "M": "Met", "V": "Val",                  # complement
    "S": "Ser", "P": "Pro", "T": "Thr", "A": "Ala", "Y": "Tyr",                  # complement
    "*": "Stop", "H": "His", "Q": "Gln", "N": "Asn", "K": "Lys",                 # complement
    "D": "Asp", "E": "Glu", "C": "Cys", "W": "Trp", "R": "Arg",                  # complement
    "G": "Gly", "U": "Sec", "?": "?"                                             # complement
}                                                                                # complement
_longAa = {                                                                      # complement
    "F": "Phenylalanine", "L": "Leucine", "I": "Isoleucine", "M": "Methionine", "V": "Valine", # complement
    "S": "Serine", "P": "Proline", "T": "Threonine", "A": "Alanine", "Y": "Tyrosine", # complement
    "*": "Stop", "H": "Histidine", "Q": "Glutamine", "N": "Asparagine", "K": "Lysine", # complement
    "D": "AsparticAcid", "E": "GlutamicAcid", "C": "Cysteine", "W": "Tryptophan", "R": "Arginine", # complement
    "G": "Glycine", "U": "Selenocysteine", "?": "?"                              # complement
}                                                                                # complement
class translate(BaseCli):                                                        # translate
    def __init__(self, length:int=0):                                            # translate
        """Translates incoming rows.

:param length: 0 for short (L), 1 for med (Leu), 2 for long (Leucine)"""         # translate
        super().__init__(); self.delim = "" if length == 0 else " "              # translate
        self.dict = [_shortAa, _medAa, _longAa][length]                          # translate
    def __ror__(self, it:Iterator[str]):                                         # translate
        super().__ror__(it)                                                      # translate
        if isinstance(it, str): it = [it]                                        # translate
        it = it | transcribe()                                                   # translate
        for line in it:                                                          # translate
            line = line.replace(" ", "")                                         # translate
            answer = ""; n = len(line)                                           # translate
            for i in range(0, n - n % 3, 3):                                     # translate
                codon = line[i:i+3].upper()                                      # translate
                answer += (self.dict[ntAa[codon]] if codon in ntAa else "?") + self.delim # translate
            yield answer                                                         # translate
class medAa(BaseCli):                                                            # medAa
    """Converts short aa sequence to medium one"""                               # medAa
    def __ror__(self, it:Iterator[str]):                                         # medAa
        if isinstance(it, str): it = [it]                                        # medAa
        for line in it:                                                          # medAa
            yield " ".join(_medAa[c] for c in line)                              # medAa
class longAa(BaseCli):                                                           # longAa
    """Converts short aa sequence to long one"""                                 # longAa
    def __ror__(self, it:Iterator[str]):                                         # longAa
        if isinstance(it, str): it = [it]                                        # longAa
        for line in it:                                                          # longAa
            yield " ".join(_longAa[c] for c in line)                             # longAa



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""This module is for all things related to atoms, molecules and their simulations"""
import k1lib
from typing import Dict, List
settings = k1lib.Settings().add("overOctet", False, "whether to allow making bonds that exceeds the octet rule")
k1lib.settings.add("mo", settings, "from k1lib.mo module")
__all__ = ["Atom", "substances", "NoFreeElectrons", "OctetFull"]
class NoFreeElectrons(RuntimeError): pass                                        # NoFreeElectrons
class OctetFull(RuntimeError): pass                                              # OctetFull
# if Atom's gDepth is smaller than this, then it means that it has not been visited # OctetFull
_depthAuto = k1lib.AutoIncrement()                                               # OctetFull
_idxAuto = k1lib.AutoIncrement()                                                 # OctetFull
class Atom:                                                                      # Atom
    """Just an atom really. Has properties, can bond to other atoms, and can
generate a :class:`System` for simulation."""                                    # Atom
                                                                                 # Atom
    def __init__(self, name:str, atomicN:int, massN:float, valenceE:int, radius:List[float]=[], octetE:int=8): # Atom
        """Creates a new atom. Not intended to be used by the end user. If you
wish to get a new atom, just do stuff like this::

    c1 = mo.C
    c2 = mo.C
    c1 == c2 # returns False, demonstrating that these are different atoms

If you wish to register new substances with the module, you can do this::

    genF = lambda: Atom(...)
    mo.registerSubstance("elementName", genF)
    mo.elementName # should executes `genF` and returns

:param name: element name (eg. "C")
:param atomicN: atomic number (eg. 6)
:param massN: atomic mass in g/mol (eg. 12)
:param valenceE: how many valence electrons initially?
:param radius: covalent radiuses (in pm) for single, double and triple bonds
:param octetE: how many electrons in a full octet? Default 8, but can be 2 for H and He""" # Atom
        self.name = name; self.atomicN = atomicN; self.massN = massN             # Atom
        self.ogValenceE = valenceE # original                                    # Atom
        self.valenceE = valenceE; self.octetE = octetE; self.radius = radius     # Atom
        self._bonds = [] # list of Atoms this Atom is bonded to                  # Atom
        self.gDepth = -1 # graph depth, for graph traversal stuff. Values will be updated from _depthAuto # Atom
        self.idx = f"A{_idxAuto()}" # unique value for Atoms everywhere          # Atom
        # contracts:                                                             # Atom
        # - valenceE = eClouds * 2 + freeE + len(bonds) * 2                      # Atom
        # - valenceE <= octetE. "<" happens when octet not full                  # Atom
        # can only form a new bond if freeE >= 1. Can dec eClouds to inc freeE   # Atom
        if name != "_e":                                                         # Atom
            self.eClouds = []; self.freeE = valenceE % 2                         # Atom
            for i in range(valenceE//2): self.eClouds.append(mo._e)              # Atom
        else: self.eClouds = []; self.freeE = 0                                  # Atom
    @property                                                                    # Atom
    def bonds(self):                                                             # Atom
        """List of Atoms bonded to this Atom"""                                  # Atom
        return self._bonds                                                       # Atom
    @bonds.setter                                                                # Atom
    def bonds(self, v): self._bonds = v                                          # Atom
    @property                                                                    # Atom
    def nonHBonds(self) -> List["Atom"]:                                         # Atom
        """All atoms this atom is bonded to, minus the Hydrogens."""             # Atom
        return [a for a in self.bonds if a.name != "H"]                          # Atom
    @property                                                                    # Atom
    def HBonds(self) -> List["Atom"]:                                            # Atom
        """All hydrogens this atom is bonded to."""                              # Atom
        return [a for a in self.bonds if a.name == "H"]                          # Atom
    @property                                                                    # Atom
    def uniqueBonds(self) -> List["Atom"]:                                       # Atom
        """All unique bonds. Meaning, if there's a double bond, only return 1
atom, not 2."""                                                                  # Atom
        return list(set(self.bonds))                                             # Atom
    @property                                                                    # Atom
    def uniqueNonHBonds(self) -> List["Atom"]:                                   # Atom
        """All unique non Hydrogen bonds."""                                     # Atom
        return list(set(self.nonHBonds))                                         # Atom
    def nBonds(self, atom:"Atom"):                                               # Atom
        """Get number of bonds between this and another atom."""                 # Atom
        return len([bond for bond in self.bonds if bond == atom])                # Atom
    @property                                                                    # Atom
    def availableBonds(self) -> int:                                             # Atom
        """Available bonds. This includes electron clouds, radical electrons, and
Hydrogen bonds."""                                                               # Atom
        return len(self.eClouds) * 2 + self.freeE + len([a for a in self.bonds if a.name == "H"]) # Atom
    def __repr__(self):                                                          # Atom
        return f"""<Atom {self.name} ({self.atomicN}), {len(self.bonds)} bonds, {self.valenceE}/{self.octetE} valence electrons, {len(self.eClouds)} electron clouds, {self.freeE} free (radical) electrons>""" # Atom
@k1lib.patch(Atom)                                                               # Atom
def _show(self, g=None, gDepth=-1, H:bool=True, GVKwargs={}):                    # _show
    self.gDepth = gDepth                                                         # _show
    if not H:                                                                    # _show
        nH = len(self.HBonds); nH = "" if nH==0 else ("H" if nH == 1 else f"H{nH}") # _show
        g.node(self.idx, f"{self.name}{nH}", **GVKwargs)                         # _show
    else: g.node(self.idx, self.name, **GVKwargs)                                # _show
    for atom in self.bonds:                                                      # _show
        if atom.gDepth >= gDepth or (not H and atom.name == "H"): continue       # _show
        # all this complexity just to determine arrow direction                  # _show
        d1 = (self.nonHBonds[0] == atom) if len(self.nonHBonds) > 0 else False   # _show
        d2 = (atom.nonHBonds[0] == self) if len(atom.nonHBonds) > 0 else False   # _show
        if d1 and d2: g(self.idx, atom.idx, dir="both")                          # _show
        elif d1: g(self.idx, atom.idx)                                           # _show
        elif d2: g(atom.idx, self.idx)                                           # _show
        else: g(self.idx, atom.idx, arrowhead="none")                            # _show
    if H: [atom._show(g, gDepth, H) for atom in self.bonds if atom.gDepth < gDepth] # _show
    else: [atom._show(g, gDepth, H) for atom in self.nonHBonds if atom.gDepth < gDepth] # _show
@k1lib.patch(Atom)                                                               # _show
def show(self, H:bool=True):                                                     # show
    """Show the molecule graph this atom is a part of. Meant for debugging
simple substances only, as graphs of big molecules look unwieldy. This also
highlights the current :class:`Atom`, and each bond is an arrow, indicating
where :meth:`next` will go next.

:param H: whether to display hydrogens as separate atoms, or bunched into the main atom""" # show
    g = k1lib.digraph(); self._show(g, _depthAuto(), H, {"style": "filled"}); return g # show
@k1lib.patch(Atom)                                                               # show
def _addFreeE(self, amt:int=1):                                                  # _addFreeE
    """Adds free electron to atom."""                                            # _addFreeE
    if amt > 1: [self._addFreeE() for i in range(amt)]                           # _addFreeE
    self.freeE += 1                                                              # _addFreeE
    if self.freeE >= 2: self.eClouds.append(mo._e); self.freeE -= 2              # _addFreeE
@k1lib.patch(Atom)                                                               # _addFreeE
def _subFreeE(self, amt:int=1) -> bool:                                          # _subFreeE
    """Tries to use ``amt`` free electrons. Returns successful or not."""        # _subFreeE
    if amt > 1: [self._subFreeE() for i in range(amt)]                           # _subFreeE
    elif self.freeE > 0: self.freeE -= 1                                         # _subFreeE
    elif len(self.eClouds) > 0:                                                  # _subFreeE
        self.freeE += 1; self.eClouds.pop()                                      # _subFreeE
    else: raise RuntimeError(f"Can't give away any more free electrons on atom {self.name}!") # _subFreeE
@k1lib.patch(Atom)                                                               # _subFreeE
def _makeRoom(self, nBonds:int):                                                 # _makeRoom
    """Tries to remove bonds with Hydrogen to make room for ``nBonds`` more bonds.""" # _makeRoom
    nBondsToRemove = self.valenceE + nBonds - self.octetE                        # _makeRoom
    if nBondsToRemove > 0:                                                       # _makeRoom
        Hs = [bond for bond in self.bonds if bond.name == "H"]                   # _makeRoom
        if len(Hs) >= nBondsToRemove:                                            # _makeRoom
            for i in range(nBondsToRemove): self.removeBond(Hs[i])               # _makeRoom
        elif not settings.overOctet:                                             # _makeRoom
            ans = input(f"Can't remove Hydrogen bonds to make room for new bond! Do you want to do anyway (y/n): ") # _makeRoom
            print("Btw, you can auto accept this by doing `settings.mo.overOctet = True`") # _makeRoom
            if ans.lower()[0] != "y": raise OctetFull("Stopping...")             # _makeRoom
    availableE = len(self.eClouds) * 2 + self.freeE                              # _makeRoom
    if availableE < nBonds: raise NoFreeElectrons(f"Can't make room for {nBonds} new bonds on {self.name}. Only {availableE} electrons left for bonds!") # _makeRoom
@k1lib.patch(Atom)                                                               # _makeRoom
def __call__(self, atom:Atom, nBonds:int=1, main=False) -> Atom:                 # __call__
    """Forms a bond with another atom. If valence electrons are full, will
attempt to disconnect Hydrogens from self to make room.

:param bond: number of bonds. 2 for double, 3 for triple
:param main: whether to put this bond in front of existing bonds, to
    signify the "main" chain, so that it works well with :meth:`next`
:return: self"""                                                                 # __call__
    self._makeRoom(nBonds); atom._makeRoom(nBonds)                               # __call__
    if main: self.bonds = [atom] * nBonds + self.bonds                           # __call__
    else: self.bonds += [atom] * nBonds                                          # __call__
    atom.bonds += [self] * nBonds                                                # __call__
    self.valenceE += nBonds; self._subFreeE(nBonds)                              # __call__
    atom.valenceE += nBonds; atom._subFreeE(nBonds)                              # __call__
    return self                                                                  # __call__
@k1lib.patch(Atom)                                                               # __call__
def bond(self, atom:Atom, nBonds:int=1, main=False) -> Atom:                     # bond
    """Like :meth:`__call__`, but returns the atom passed in instead, so you
can form the main loop quickly."""                                               # bond
    self(atom, nBonds, main); return atom                                        # bond
@k1lib.patch(Atom)                                                               # bond
def main(self, atom:Atom, nBonds:int=1) -> Atom:                                 # main
    """Like :meth:`bond`, but with ``main`` param defaulted to True."""          # main
    return self.bond(atom, nBonds, True)                                         # main
@k1lib.patch(Atom)                                                               # main
def removeBond(self, atom:"Atom"):                                               # removeBond
    """Removes all bonds between this and another atom"""                        # removeBond
    nBonds = self.nBonds(atom)                                                   # removeBond
    self.bonds = [bond for bond in self.bonds if bond != atom]                   # removeBond
    self.valenceE -= nBonds; self._addFreeE(nBonds)                              # removeBond
    atom.bonds = [bond for bond in atom.bonds if bond != self]                   # removeBond
    atom.valenceE -= nBonds; atom._addFreeE(nBonds)                              # removeBond
@k1lib.patch(Atom, "next")                                                       # removeBond
def _next(self, offset=0, times:int=1) -> "Atom":                                # _next
    """Returns the next atom bonded to this. Tries to avoid going into Hydrogens.
This is the main way to navigate around the molecule.

You kinda have to make sure that your molecule's bonding order is appropriate by
choosing between :meth:`bond` and :meth:`main`. Check the bonding order with
:meth:`show`.

:param offset: if there are multiple non-Hydrogen atoms, which ones should I pick?
:param times: how many times do you want to chain ``.next()``?"""                # _next
    if times < 0: raise RuntimeError("Can't do .next() with negative `times`")   # _next
    if times == 0: return self                                                   # _next
    atoms = self.nonHBonds + self.HBonds                                         # _next
    if len(atoms) == 0: return None                                              # _next
    _next = atoms[offset]                                                        # _next
    if times == 1: return _next                                                  # _next
    else: return _next.next(offset, times-1)                                     # _next
@k1lib.patch(Atom)                                                               # _next
def nexts(self, atoms:int=2) -> List[Atom]:                                      # nexts
    """Kinda like :meth:`next`, but fetches multiple atoms on the backbone.
Example::

    c1, c2 = mo.CH4(mo.CH4).nexts()"""                                           # nexts
    if atoms < 1: raise RuntimeError(f"Zero or negative ({atoms}) number of atoms does not make sense!") # nexts
    if atoms == 1: return [self]                                                 # nexts
    return [self, *(self.next().nexts(atoms-1))]                                 # nexts
empiricalOrder = ["C", "H", "O", "N"]                                            # nexts
def em1(e:str, n:int):                                                           # em1
    if n == 1: return e                                                          # em1
    else: return f"{e}{n}"                                                       # em1
@k1lib.patch(Atom)                                                               # em1
def _empirical(self, d:Dict[str, int], gDepth:int):                              # _empirical
    if self.gDepth >= gDepth: return                                             # _empirical
    self.gDepth = gDepth; d[self.name] += 1                                      # _empirical
    for atom in self.bonds: atom._empirical(d, gDepth)                           # _empirical
@k1lib.patch(Atom)                                                               # _empirical
def empirical(self) -> str:                                                      # empirical
    """Returns an empirical formula for the molecule this :class:`Atom` is attached to.""" # empirical
    d = k1lib.Object().withAutoDeclare(lambda: 0)                                # empirical
    self._empirical(d, _depthAuto()); answer = ""                                # empirical
    for e in empiricalOrder:                                                     # empirical
        if e in d: answer += em1(e,d[e]); del d[e]                               # empirical
    for e in d.state.keys(): answer += em1(e,d[e])                               # empirical
    return answer                                                                # empirical
@k1lib.patch(Atom)                                                               # empirical
def _atoms(self, l, gDepth):                                                     # _atoms
    if self.gDepth >= gDepth: return                                             # _atoms
    self.gDepth = gDepth; l.append(self)                                         # _atoms
    for atom in self.bonds: atom._atoms(l, gDepth)                               # _atoms
@k1lib.patch(Atom)                                                               # _atoms
def atoms(self) -> List[Atom]:                                                   # atoms
    """Returns a list of Atoms in the molecule this specific Atom is attached to.""" # atoms
    l = []; self._atoms(l, _depthAuto()); return l                               # atoms
@k1lib.patch(Atom, "endChain")                                                   # atoms
@property                                                                        # atoms
def endChain(a) -> Atom:                                                         # endChain
    """Do a bunch of .next() until reached the end of the carbon chain.
Example::

    c1 = mo.alcohol(3, 1)
    c3 = c1.endChain
    c3(mo.NH3)
    c1.show() # displays in cell"""                                              # endChain
    lastA = None                                                                 # endChain
    for i in range(200): # for loop to prevent infinite recursion                # endChain
        nextA = a.next()                                                         # endChain
        if nextA == lastA: return a                                              # endChain
        lastA = a; a = nextA                                                     # endChain
@k1lib.patch(Atom)                                                               # endChain
def moveLastCTo2ndC(a:Atom) -> Atom:                                             # moveLastCTo2ndC
    """Move last carbon to 2nd carbon. Useful in constructing iso- and tert-.""" # moveLastCTo2ndC
    end = a.endChain; nearEnd = end.next()                                       # moveLastCTo2ndC
    end.removeBond(nearEnd); nearEnd(mo.H); a.next()(mo.CH4); return a           # moveLastCTo2ndC
_a = {} # dict of atoms, which will be used to patch the entire module           # moveLastCTo2ndC
class _Mo:                                                                       # _Mo
    def __init__(self): self._MoWrap_dirs = []                                   # _Mo
    def registerSubstance(self, name:str, _f):                                   # _Mo
        setattr(_Mo, name, property(lambda self: _f()))                          # _Mo
        self._MoWrap_dirs.append(name)                                           # _Mo
    def __dir__(self):                                                           # _Mo
        return super().__dir__() + self._MoWrap_dirs                             # _Mo
    pass                                                                         # _Mo
mo = _Mo() # internal convenience object so that I can use the same style as the module # _Mo
def _atom(name, *args, **kwargs):                                                # _atom
    _a[name] = f = lambda: Atom(name, *args, **kwargs)                           # _atom
    mo.registerSubstance(name, f)                                                # _atom
def substances() -> List[str]:                                                   # substances
    """Get a list of builtin substances. To register new substances, check over
:class:`Atom`."""                                                                # substances
    return [k for k in _a.keys() if not k.startswith("_")]                       # substances
# covalent radius taken from (Pyykko & Atsumi) https://chem.libretexts.org/@api/deki/pages/2182/pdf/A3%253A%2bCovalent%2bRadii.pdf?stylesheet=default # substances
_atom("_e", 0,   0.1,    0, [25]) # electron cloud, for internal use             # substances
_atom("H",  1,   1.008,  1, [32], octetE=2)                                      # substances
_atom("Li", 3,   6.94,   1, [133, 124])                                          # substances
_atom("Be", 4,   9.0122, 2, [102, 90, 85])                                       # substances
_atom("B",  5,  10.81,   3, [85,  78, 73])                                       # substances
_atom("C",  6,  12.011,  4, [75,  67, 60])                                       # substances
_atom("N",  7,  14.007,  5, [71,  60, 54])                                       # substances
_atom("O",  8,  15.999,  6, [63,  57, 53])                                       # substances
_atom("F",  9,  18.998,  7, [64,  59, 53])                                       # substances
_atom("Na", 11, 22.990, 1, [155, 160])                                           # substances
_atom("Mg", 12, 24.305, 2, [139, 132, 127])                                      # substances
_atom("Al", 13, 26.982, 3, [126, 113, 111])                                      # substances
_atom("Si", 14, 28.085, 4, [116, 107, 102])                                      # substances
_atom("P",  15, 30.974, 5, [111, 102, 94])                                       # substances
_atom("S",  16, 32.06,  6, [103, 94,  95])                                       # substances
_atom("Cl", 17, 35.45,  7, [99,  95,  93])                                       # substances
_atom("K",  19,  39.098, 1, [196, 193])                                          # substances
_atom("Ca", 20,  40.078, 2, [171, 147, 133])                                     # substances
_atom("Ga", 31,  69.723, 3, [124, 117, 121])                                     # substances
_atom("Ge", 32,  72.630, 4, [121, 111, 114])                                     # substances
_atom("As", 33,  74.922, 5, [121, 114, 106])                                     # substances
_atom("Se", 34,  78.971, 6, [116, 107, 107])                                     # substances
_atom("Br", 35,  79.904, 7, [114, 109, 110])                                     # substances
_atom("I",  53, 126.9,   7, [133, 129, 125])                                     # substances



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import logging, warnings, os, time, re, json, k1lib, importlib, urllib.parse, math, base64, dill, inspect, threading, datetime
import numpy as np
from typing import Any, List, Union, Tuple, Iterator, Dict
from functools import partial
try: import torch; hasTorch = True
except: hasTorch = False
__all__ = ["_docsUrl", "isNumeric",
           "patch", "cron", "wrapMod", "wraps", "squeeze", "raiseEx",
           "numDigits", "limitLines",
           "limitChars", "showLog", "cleanDiv",
           "beep", "beepOnAvailable", "dontWrap",
           "debounce", "scaleSvg", "now", "pushNotification", "dep", "depCli", "ticks", "digraph", "graph",
           "encode", "decode", "hash", "resolve", "config"]
_docsUrl = "https://k1lib.com"
def isNumeric(x) -> bool:                                                        # isNumeric
    """Returns whether object is actually a number"""                            # isNumeric
    return isinstance(x, (int, float, np.number))                                # isNumeric
def patch(_class:type, name:str=None, docs:Union[str, Any]=None, static=False):  # patch
    """Patches a function to a class/object.

:param _class: object to patch function. Can also be a type
:param name: name of patched function, if different from current
:param docs: docs of patched function. Can be object with defined __doc__ attr
:param static: whether to wrap this inside :class:`staticmethod` or not
:return: modified function just before patching

Intended to be used like this::

    class A:
        def methA(self):
            return "inside methA"

    @k1lib.patch(A)
    def methB(self):
        return "inside methB"

    a = A()
    a.methB() # returns "inside methB"

You can do ``@property`` attributes like this::

    class A: pass

    @k1lib.patch(A, "propC")
    @property
    def propC(self): return self._propC

    @k1lib.patch(A, "propC")
    @propC.setter
    def propC(self, value): self._propC = value

    a = A(); a.propC = "abc"
    a.propC # returns "abc"

The attribute name unfortunately has to be explicitly declared, as I can't
really find a way to extract the original name. You can also do static
methods like this::

    class A: pass

    @k1lib.patch(A, static=True)
    def staticD(arg1): return arg1

    A.staticD("def") # returns "def"
"""                                                                              # patch
    def inner(function):                                                         # patch
        _docs = docs                                                             # patch
        if _docs is not None and not isinstance(_docs, str): _docs = _docs.__doc__ # patch
        _docs = _docs or function.__doc__ or _class.__doc__                      # patch
        _name = name or function.__qualname__.split(".")[-1]                     # patch
                                                                                 # patch
        _function = staticmethod(function) if static else function               # patch
        _function.__doc__ = _docs; setattr(_class, _name, _function); return _function # patch
    return inner                                                                 # patch
def cron(f):                                                                     # cron
    """Sets up a cron job in another thread, running the decorated
function whenever ``f`` goes from False to True. Example::

    @k1.cron(lambda minute: minute == 0)
    def f1(): # runs every hour
        ... # do some stuff

    @k1.cron(lambda second: second % 5 == 0)
    def f2(): # runs every 5 seconds
        ... # do some stuff

So, the first function will run every hour, and the second function will
run every 5 seconds. Pretty straightforward. The timing function ``f`` can
be as complicated as you want, but it can only accept the following parameters:

- year
- month: 1-12
- day: 1-31
- weekday: 0-6, 0 for Monday
- hour: 0-23
- minute: 0-59
- second: 0-59"""                                                                # cron
    def inner(func):                                                             # cron
        args = list(inspect.signature(f).parameters.keys())                      # cron
        s = {"year", "month", "day", "weekday", "hour", "minute", "second"}      # cron
        for arg in args:                                                         # cron
            if arg not in s: raise Exception(f"Unknown argument {arg}. Only (year, month, day, weekday, hour, minute, seconds) are allowed") # cron
        def startLoop():                                                         # cron
            last = False; this = last                                            # cron
            while True:                                                          # cron
                a = datetime.datetime.now()                                      # cron
                now = {"year": a.year, "month": a.month, "day": a.day, "weekday": a.weekday(), "hour": a.hour, "minute": a.minute, "second": a.second} # cron
                this = f(*[now[e] for e in args])                                # cron
                if not last and this: func()                                     # cron
                last = this; time.sleep(0.5)                                     # cron
        threading.Thread(target=startLoop).start()                               # cron
        return func                                                              # cron
    return inner                                                                 # cron
class wrapMod:                                                                   # wrapMod
    def __init__(self, m, moduleName=None):                                      # wrapMod
        """Wraps around a module, and only suggest symbols in __all__ list
defined inside the module. Example::

    from . import randomModule
    randomModule = wrapMod(randomModule)

:param m: the imported module
:param moduleName: optional new module name for elements (their __module__ attr)""" # wrapMod
        self._init(m, moduleName)                                                # wrapMod
    def _init(self, m, moduleName):                                              # wrapMod
        if moduleName is not None:                                               # wrapMod
            for v in m.__dict__.values():                                        # wrapMod
                v.__module__ = moduleName                                        # wrapMod
        self._wrapMod_moduleName = moduleName; self._wrapMod_m = m               # wrapMod
        self.__dict__.update(m.__dict__)                                         # wrapMod
        self._wrapMod_extraDirs = []                                             # wrapMod
    def __dir__(self): return self._wrapMod_m.__all__ + self._wrapMod_extraDirs  # wrapMod
    def __str__(self): return str(self._wrapMod_m)                               # wrapMod
    def __getstate__(self): return {"m": self._wrapMod_m, "name": self._wrapMod_moduleName} # wrapMod
    def __setstate__(self, d): self._init(d["m"], d["name"])                     # wrapMod
    def __repr__(self): return str(self)                                         # wrapMod
def wraps(ogF):                                                                  # wraps
    """Kinda like :meth:`functools.wraps`, but don't update __annotations__."""  # wraps
    def inner(f):                                                                # wraps
        f.__doc__ = ogF.__doc__                                                  # wraps
        f.__name__ = ogF.__name__                                                # wraps
        f.__qualname__ = ogF.__qualname__                                        # wraps
        f.__module__ = ogF.__module__                                            # wraps
        return f                                                                 # wraps
    return inner                                                                 # wraps
def squeeze(_list:Union[list, tuple, Any], hard=False):                          # squeeze
    """If list only has 1 element, returns that element, else returns original
list.

:param hard: If True, then if list/tuple, filters out None, and takes the first
    element out even if that list/tuple has more than 1 element"""               # squeeze
    if isinstance(_list, (tuple, list)):                                         # squeeze
        if hard: return [e for e in _list if e != None and e != ""][0]           # squeeze
        elif len(_list) == 1: return _list[0]                                    # squeeze
    if hasTorch and isinstance(_list, torch.Tensor): return _list.squeeze()      # squeeze
    return _list                                                                 # squeeze
def raiseEx(ex:Exception):                                                       # raiseEx
    """Raises a specific exception. May be useful in lambdas"""                  # raiseEx
    raise ex                                                                     # raiseEx
def numDigits(num) -> int:                                                       # numDigits
    """Get the number of digits/characters of this number/object"""              # numDigits
    return len(f"{num}")                                                         # numDigits
def limitLines(s:str, limit:int=10) -> str:                                      # limitLines
    """If input string is too long, truncates it and adds ellipsis"""            # limitLines
    splits = s.split("\n")                                                       # limitLines
    if len(splits) > limit: return "\n".join(splits[:limit]) + "\n....."         # limitLines
    else: return s                                                               # limitLines
def limitChars(s:str, limit:int=50):                                             # limitChars
    """If input string is too long, truncates to first `limit` characters of the first line""" # limitChars
    if s is None: return ""                                                      # limitChars
    s = f"{s}".split("\n")[0]                                                    # limitChars
    return s[:limit-3] + "..." if len(s) > limit else s                          # limitChars
def showLog(loggerName:str="", level:int=logging.DEBUG):                         # showLog
    """Prints out logs of a particular logger at a particular level"""           # showLog
    logger = logging.getLogger(loggerName); logger.setLevel(level)               # showLog
    sh = logging.StreamHandler(); sh.setLevel(logging.DEBUG); logger.addHandler(sh) # showLog
def cleanDiv(_list:List[float], total:int) -> List[int]:                         # cleanDiv
    """Cleanly divides :total: into int segments with weights specified by
:_list:. Looks like this: ``(_list / _list.sum()) * total``, but
everything is an ``int``"""                                                      # cleanDiv
    _list = np.array(_list)                                                      # cleanDiv
    _list = (_list*total/_list.sum()).astype(int)                                # cleanDiv
    _list[-1] = total - _list[:-1].sum()                                         # cleanDiv
    return _list                                                                 # cleanDiv
def beep(seconds=0.3):                                                           # beep
    """Plays a beeping sound, may be useful as notification for long-running tasks""" # beep
    try: import IPython; IPython.core.display.display_html(IPython.display.HTML(f"""<script>(new Audio('data:audio/wav;base64,UklGRl9vT19XQVZFZm10IBAAAAABAAEAQB8AAEAfAAABAAgAZGF0YU'+Array(Math.round(3.3333e3*{seconds})).join(123))).play();</script>""")); # beep
    except: os.system("printf '\a'")                                             # beep
def beepOnAvailable(url:str, timeout=5, **kwargs):                               # beepOnAvailable
    """Tries to connect with a url repeatedly, and if successful, plays
a beep sound"""                                                                  # beepOnAvailable
    import requests                                                              # beepOnAvailable
    try:                                                                         # beepOnAvailable
        while True:                                                              # beepOnAvailable
            time.sleep(1); successful = False                                    # beepOnAvailable
            try:                                                                 # beepOnAvailable
                if requests.get(url, timeout=timeout, **kwargs).ok:              # beepOnAvailable
                    successful = True                                            # beepOnAvailable
            except: pass                                                         # beepOnAvailable
            if successful:                                                       # beepOnAvailable
                beep(); break                                                    # beepOnAvailable
    except KeyboardInterrupt: print("Still not available")                       # beepOnAvailable
def dontWrap():                                                                  # dontWrap
    """Don't wrap horizontally when in a notebook. Normally, if you're
displaying something long, like the output of ``print('a'*1000)`` in a notebook,
it will display it in multiple lines. This may be undesirable, so this solves
that by displaying some HTML with css styles so that the notebook doesn't wrap.""" # dontWrap
    try:                                                                         # dontWrap
        from IPython.core.display import display, HTML                           # dontWrap
        display(HTML("""<style>
    div.jp-OutputArea-output pre {white-space: pre;}
    div.output_area pre {white-space: pre;}
    div.CodeMirror > div.highlight {overflow-y: auto;}
</style>"""))                                                                    # dontWrap
    except: pass#print("Can't run dontWrap()")                                   # dontWrap
import asyncio, functools                                                        # dontWrap
from threading import Timer as ThreadingTimer                                    # dontWrap
class AsyncTimer: # rename if want to use                                        # AsyncTimer
    def __init__(self, timeout, callback):                                       # AsyncTimer
        self._timeout = timeout; self._callback = callback                       # AsyncTimer
    async def _job(self):                                                        # AsyncTimer
        await asyncio.sleep(self._timeout); self._callback()                     # AsyncTimer
    def start(self): self._task = asyncio.ensure_future(self._job())             # AsyncTimer
    def cancel(self): self._task.cancel()                                        # AsyncTimer
def debounce(wait, threading=False):                                             # debounce
    """Decorator that will postpone a function's execution until after
``wait`` seconds have elapsed since the last time it was invoked. Taken
from `ipywidgets <https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20Events.html>`_.
Example::

    import k1lib, time; value = 0

    @k1lib.debounce(0.5, True)
    def f(x): global value; value = x**2

    f(2); time.sleep(0.3); f(3)
    print(value) # prints "0"
    time.sleep(0.7)
    print(value) # prints "9"

:param wait: wait time in seconds
:param threading: if True, use multiple threads, else just use async stuff"""    # debounce
    Timer = ThreadingTimer if threading else AsyncTimer                          # debounce
    def decorator(fn):                                                           # debounce
        timer = None                                                             # debounce
        def debounced(*args, **kwargs):                                          # debounce
            nonlocal timer                                                       # debounce
            if timer is not None: timer.cancel()                                 # debounce
            timer = Timer(wait, lambda: fn(*args, **kwargs))                     # debounce
            timer.start()                                                        # debounce
        functools.update_wrapper(debounced, fn); return debounced                # debounce
    return decorator                                                             # debounce
def scaleSvg(svg:str, scale:float=None) -> str:                                  # scaleSvg
    """Scales an svg xml string by some amount."""                               # scaleSvg
    if scale is None: scale = k1lib.settings.svgScale                            # scaleSvg
    wS = w = re.findall("width=\"\d*pt\"", svg)[0]                               # scaleSvg
    hS = w = re.findall("height=\"\d*pt\"", svg)[0]                              # scaleSvg
    w = int(int(re.findall("\d+", wS)[0])*scale)                                 # scaleSvg
    h = int(int(re.findall("\d+", hS)[0])*scale)                                 # scaleSvg
    svg = re.sub(wS, f'width="{w}pt"', svg)                                      # scaleSvg
    svg = re.sub(hS, f'height="{h}pt"', svg)                                     # scaleSvg
    return svg                                                                   # scaleSvg
import datetime                                                                  # scaleSvg
def now():                                                                       # now
    """Convenience function for returning a simple time
string, with timezone and whatnot."""                                            # now
    return datetime.datetime.now().astimezone().isoformat()                      # now
now()                                                                            # now
def pushNotification(title="Some title", content="Some content", url="https://k1lib.com"): # pushNotification
    """Sends push notification to your device.

Setting things up:
- Download this app: https://play.google.com/store/apps/details?id=net.xdroid.pn
- Set the `settings.pushNotificationKey` key obtained from the app. Example key: `k-967fe9...`
- Alternatively, set the environment variable `k1lib_pushNotificationKey` instead
- Run the function as usual"""                                                   # pushNotification
    import requests                                                              # pushNotification
    key = k1lib.settings.pushNotificationKey                                     # pushNotification
    requests.get("http://xdroid.net/api/message?" + urllib.parse.urlencode({'k': key, 't': title, 'c': content, 'u': url})) # pushNotification
    print("Pushed!")                                                             # pushNotification
class Dependency:                                                                # Dependency
    def __init__(self, s, alt:str=None, url:str=None): self.s = s; self.alt = alt; self.url = url # Dependency
    def __getattr__(self, attr):                                                 # Dependency
        after = f"More information is available on {self.url}" if self.url else f"" # Dependency
        raise Exception(f"Python package `{self.alt or self.s}` not found. Please install it. {after}") # Dependency
def dep(s, alt:str=None, url:str=None):                                          # dep
    """Imports a potentially unavailable package
Example::

    graphviz = k1.dep("graphviz")

    # executes as normal, if graphviz is available, else throws an error
    g = graphviz.Digraph()

I don't imagine this would be useful for everyday use though.
This is mainly for writing this library, so that it can use
optional dependencies.

:param s: name of the package. Can be nested, like `matplotlib.pyplot`
:param alt: (optional) name of the package to display in error message if package is not found
:param url: (optional) url of the package's website, so that they know where to get official docs""" # dep
    try: return importlib.import_module(s)                                       # dep
    except: return Dependency(s, alt, url)                                       # dep
def depCli(prog="ls", insCmd=None, website=None):                                # depCli
    """Checks whether a particular command-line program is installed or not.
Example::

    k1.depCli("vim")

This will throw an error if that program is not found

:param prog: program name
:param insCmd: (optional) installation command, could be "apt install vim"
:param website: (optional) website that the user can learn more about the program""" # depCli
    if None | k1lib.cli.cmd(("where " if os.name == "nt" else "which ") + prog) | k1lib.cli.shape(0) == 0: # depCli
        insS = f" by executing '{insCmd}'" if insCmd else ""                     # depCli
        webS = f". Visit {website} to learn more about the program" if website else "" # depCli
        raise Exception(f"Command line program '{prog}' not found in PATH. Please install it{insS}{webS}") # depCli
    return True                                                                  # depCli
tickCheckpoints = np.array([1, 2, 2.5, 5, 10])/10                                # depCli
def ticks(x:float, y:float, rounding:int=6):                                     # ticks
    """Get tick locations in a plot that look reasonable.
Example::

    ticks(-5, 40)     # returns [-10.0, -5.0, 0.0, 5.0, 10.0, 15.0, 20.0, 25.0, 30.0, 35.0, 40.0, 45.0]
    ticks(0.05, 0.07) # returns [0.05, 0.0525, 0.055, 0.0575, 0.06, 0.0625, 0.065, 0.0675, 0.07, 0.0725]
    ticks(-5, 5)      # returns [-6.0, -5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0]

So essentially, whenever you try to plot something, you want both the x and y axis
to not have too many lines, and that the tick values snap to a nice number.

Normally you don't have to do this, as matplotlib does this automatically behind
the scenes, but sometimes you need to implement plotting again, in strange
situations, so this function comes in handy

:param x: start of interval
:param y: end of interval
:param rounding: internally, it rounds tick values to this number of
    digits, to fix tiny float overflows that make numbers ugly. So
    you can disable it if you're working with really small numbers"""            # ticks
    cli = k1lib.cli; ld = math.log10(y-x); scale = ld-ld%1 # log delta           # ticks
    sel = ((ld%1 - tickCheckpoints)**2).argmin() # selecting the best fit tick checkpoint # ticks
    interval = 10**scale*tickCheckpoints[sel] # interval between ticks           # ticks
    seed = int(y/10**scale)*10**scale # seed tick                                # ticks
    f1 = cli.applySerial(cli.op()-interval) | cli.breakIf(cli.op()<x-interval)   # ticks
    f2 = cli.applySerial(cli.op()+interval) | cli.breakIf(cli.op()>y+interval)   # ticks
    # finally, use the seed to expand in both directions to get all the ticks    # ticks
    return seed | f1 & f2 | cli.joinStreams() | cli.aS(set) | cli.sort(None) | cli.apply(round, ndigits=rounding) | cli.deref() | cli.aS(np.array) # ticks
try:                                                                             # ticks
    import graphviz                                                              # ticks
    def digraph():                                                               # ticks
        """Convenience method for creating a new graphviz Digraph.
Example::

    g = k1lib.graph()
    g("a", "b", "c")
    g # displays arrows from "a" to "b" and "a" to "c"
"""                                                                              # ticks
        return graphviz.Digraph(graph_attr={"rankdir":"TB"})                     # ticks
    def graph():                                                                 # ticks
        """Convenience method for creating a new graphviz Graph. See also: :meth:`digraph`""" # ticks
        return graphviz.Graph(graph_attr={"rankdir":"TB"})                       # ticks
except ImportError:                                                              # ticks
    digraph = graph = lambda: print("Module `graphviz` not found! Please install it first, something like `pip install graphviz`") # ticks
def encode(obj:object) -> str:                                                   # encode
    """Serialize random objects into bytes, then turn those bytes to
normal strings using base64. Example::

    a = {"b": 3}
    encode(a) # returns "gASVCgAAAAAAAAB9lIwBYpRLA3Mu"
    decode(encode(a)) # returns {"b"}: 3

See also: :meth:`decode`
"""                                                                              # encode
    return base64.b64encode(dill.dumps(obj)).decode()                            # encode
def decode(s:str) -> object:                                                     # decode
    """Turns a string produced by :meth:`encode` back into a random object."""   # decode
    return dill.loads(base64.b64decode(s.encode()))                              # decode
import hashlib                                                                   # decode
def hash(msg:str) -> int:                                                        # hash
    """A universal hash function. Why not just use the builtin hash function?
Because it actually changes for different interpreter instances, it won't be
good for code that runs on multiple computers, so this is sort of like a
drop-in replacement. Although it does output an integer, don't rely on it having
the same numeric properties as a normal hash function."""                        # hash
    if not isinstance(msg, str): msg = f"{msg}"                                  # hash
    m = hashlib.sha256(); m.update(f"{msg}".encode()); return int.from_bytes(m.digest(), "big") # hash
hash(34)                                                                         # hash
import traceback                                                                 # hash
try:                                                                             # hash
    import asyncio, threading, time; from collections import deque               # hash
    _coroutineQueue = deque() # deque of (idx, coroutine)                        # hash
    _coroutineAns = dict() # Dict[idx -> coroutine ans]                          # hash
    _coroutineAutoIdx = 0                                                        # hash
    def _coroutineResolvingThread():                                             # hash
        loop = asyncio.new_event_loop()                                          # hash
        while True:                                                              # hash
            if len(_coroutineQueue) == 0: time.sleep(0.01)                       # hash
            else:                                                                # hash
                idx, coroutine = _coroutineQueue.popleft()                       # hash
                # important to recover from exceptions                           # hash
                try: ans = loop.run_until_complete(coroutine); _coroutineAns[idx] = {"type": "success", "ans": ans} # hash
                except Exception as e: _coroutineAns[idx] = {"type": "failure", "e": f"{e}", "tb": traceback.format_exc()} # hash
    threading.Thread(target=_coroutineResolvingThread, daemon=True).start()      # hash
    _resolve_err = None                                                          # hash
except Exception as e: _resolve_err = f"{e}"; _resolve_tb = traceback.format_exc() # hash
def resolve(coroutine):                                                          # resolve
    """Resolves coroutines without having to use await.
Example::

    async def f(x):
        await asyncio.sleep(1) # simulates heavy processing
        return x + 4

    k1.resolve(f(5))

This kinda feels just like ``asyncio.run(f(5))``, so why does this exist?
Here's the docstring of that method:

.. code-block:: text

    This function cannot be called when another asyncio event loop is running in the same thread.

    This function always creates a new event loop and closes it at the end. It should be used as a main entry point for asyncio programs, and should ideally only be called once.

This has more limitations that I found annoying to deal with day-to-day. I want
a function that always work, no matter my setup. So, how this function work
internally is that it spins up a new (permanent, daemon) thread, creates a new
event loop in that thread, then whenever a coroutine comes in, it runs it in
that event loop, returns, then pass control back to whatever thread that called
:meth:`resolve`. I'm sure this can still be messed up in some way, but seems
more useful than the builtin method.

This is just meant as a quick and dirty way to force resolving coroutines. Use
this sparingly, as performance won't be as good as a proper async application.
If you find yourself using this way too often, then I'd suggest reviewing how
:mod:`asyncio` works"""                                                          # resolve
    global _coroutineAutoIdx                                                     # resolve
    if _resolve_err: raise Exception(f"k1lib.resolve() not available, encoutered this error while starting up: {_resolve_err}. Traceback:\n{_resolve_tb}") # resolve
    idx = _coroutineAutoIdx; _coroutineAutoIdx += 1; _coroutineQueue.append([idx, coroutine]) # resolve
    while idx not in _coroutineAns: time.sleep(0.01)                             # resolve
    ans = _coroutineAns[idx]; del _coroutineAns[idx]                             # resolve
    if ans["type"] == "success": return ans["ans"]                               # resolve
    else: raise Exception(f"Exception occured while trying to k1lib.resolve(): {ans['e']}. Traceback:\n{ans['tb']}") # resolve
def config(s:str):                                                               # config
    """Convenience method to grab JSON config file from config.mlexps.com"""     # config
    return json.loads("".join(k1lib.cli.cat(f"https://config.mlexps.com/{s}")))  # config



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""For not very complicated loss functions"""
from ..callbacks import Callback, Callbacks, Cbs
from typing import Callable, Tuple
import k1lib, math
try: import torch; import torch.nn.functional as F; hasTorch = True
except: torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["LossF", "LossNLLCross"]
LossFSig = Callable[[Tuple[torch.Tensor, torch.Tensor]], float]
@k1lib.patch(Cbs)
@k1lib.patch(Callback.lossCls)
class LossF(Callback):                                                           # LossF
    " "                                                                          # LossF
    def __init__(self, lossF:LossFSig):                                          # LossF
        """Generic loss function.
Expected variables in :class:`~k1lib.Learner`:

- y: result of model. Auto-included in :class:`~k1lib.callbacks.core.CoreNormal`
  and :class:`~k1lib.callbacks.core.CoreRNN`.

Deposits variables into :class:`~k1lib.Learner` at checkpoint ``inLoss``:

- lossG: single float tensor value, attached to graph
- loss: lossG, but single float value

:param lossF: takes in ``(y, yb)`` and returns ``lossG``"""                      # LossF
        super().__init__()                                                       # LossF
        self.lossF = lossF                                                       # LossF
    def inLoss(self):                                                            # LossF
        self.l.lossG = self.lossF(self.l.y, self.l.yb)                           # LossF
        self.l.loss = self.l.lossG.detach().item()                               # LossF
class LossNLLCross(Callback):                                                    # LossNLLCross
    " "                                                                          # LossNLLCross
    def __init__(self, nll:bool, integrations:bool):                             # LossNLLCross
        """Adds a cross-entropy/negative-likelihood loss function.

:param nll: if True, then use :class:`torch.nn.NLLLoss`, else use :class:`torch.nn.CrossEntropyLoss`
:param integrations: whether to integrate with :class:`~k1lib.callbacks.accuracy.AccF` callback""" # LossNLLCross
        super().__init__(); self.integrations = integrations; self.ownsAccCb = False # LossNLLCross
        self.order = 11 # to make sure it's after AccF                           # LossNLLCross
        self.lossF = torch.nn.NLLLoss() if nll else torch.nn.CrossEntropyLoss()  # LossNLLCross
        self.accuracyCbTop5 = None                                               # LossNLLCross
    def attached(self): # delayed initialization, so that learner and cbs has already been attached # LossNLLCross
        if self.integrations:                                                    # LossNLLCross
            if "AccF" not in self.cbs:                                           # LossNLLCross
                self.ownsAccCb = True                                            # LossNLLCross
                self.accuracyCb = Cbs.AccF(); self.cbs.add(self.accuracyCb)      # LossNLLCross
                self.accuracyCbTop5 = Cbs.AccF(lambda y: y, lambda y, yb: (yb[:,None] == y.topk(5, dim=1).indices+0).sum(1), variable="accTop5", hookToLearner=False); self.cbs.add(self.accuracyCbTop5) # LossNLLCross
                self.cbs.add(Cbs.Accuracy("accTop5"), name="AccuracyTop5")       # LossNLLCross
            else: self.accuracyCb = self.cbs.AccF                                # LossNLLCross
    def inLoss(self):                                                            # LossNLLCross
        self.l.lossG = self.lossF(self.l.y, self.l.yb)                           # LossNLLCross
        self.l.loss = self.l.lossG.detach().item()                               # LossNLLCross
    def detach(self):                                                            # LossNLLCross
        super().detach()                                                         # LossNLLCross
        if self.accuracyCb != None:                                              # LossNLLCross
            if self.ownsAccCb: self.accuracyCb.detach()                          # LossNLLCross
            self.accuracyCb = None                                               # LossNLLCross
        if self.accuracyCbTop5 != None:                                          # LossNLLCross
            if self.ownsAccCb: self.accuracyCbTop5.detach()                      # LossNLLCross
            self.accuracyCbTop5 = None                                           # LossNLLCross
@k1lib.patch(Cbs)                                                                # LossNLLCross
@k1lib.patch(Callback.lossCls)                                                   # LossNLLCross
class LossCrossEntropy(LossNLLCross):                                            # LossCrossEntropy
    def __init__(self, integrations:bool=True):                                  # LossCrossEntropy
        """Cross entropy loss function. Deposits into :class:`~k1lib.Learner`
the same variables as in :class:`LossF`."""                                      # LossCrossEntropy
        super().__init__(False, integrations)                                    # LossCrossEntropy
@k1lib.patch(Cbs)                                                                # LossCrossEntropy
@k1lib.patch(Callback.lossCls)                                                   # LossCrossEntropy
class LossNLL(LossNLLCross):                                                     # LossNLL
    def __init__(self, integrations:bool=True):                                  # LossNLL
        """Negative log loss function. Deposits into :class:`~k1lib.Learner`
the same variables as in :class:`LossF`."""                                      # LossNLL
        super().__init__(True, integrations)                                     # LossNLL



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import k1lib as k1; import k1lib.cli as cli
from k1lib.cli import BaseCli
import os
__all__ = ["create", "apply", "filt", "select", "flatten", "pretty", "display", "search", "walk"]
settings = k1.settings.cli
kSett = k1.Settings().add("tok", "\ue000", "Special token for internal processing")
settings.add("ktree", kSett, "cli.ktree module settings");
def _mktree(depth):                                                              # _mktree
    if depth <= 0: return cli.iden()                                             # _mktree
    return cli.groupBy(0, True) | cli.apply(_mktree(depth-1), 1)                 # _mktree
class mktree1(BaseCli):                                                          # mktree1
    def __ror__(self, it): it = it | cli.deref(2); return (it | _mktree(len(it[0])))# if len(it) > 0 else [] # mktree1
def prune1(x): # eliminating empty elements. For some reason, _mktree() creates an empty element at each leaf. No idea why and too lazy to dig in further # prune1
    for elem in x:                                                               # prune1
        if len(elem) == 0: continue                                              # prune1
        s, children = elem; yield [s, prune1(children)]                          # prune1
def prune2(x) -> "[children, hasToken]": # trying to grab whether there is a branch that ends at the parent or not # prune2
    hasToken = False; ans = []                                                   # prune2
    for s, children in x:                                                        # prune2
        children, hT = prune2(children)                                          # prune2
        hasToken = hasToken or s == kSett.tok                                    # prune2
        if len(children) == 0: hT = 1                                            # prune2
        ans.append([s, children, hT+0])                                          # prune2
    return ans, hasToken                                                         # prune2
def prune3(x, depth=0): # trying to eliminate the "virtual" elements tagged with the token # prune3
    ans = []                                                                     # prune3
    for s, children, num in x:                                                   # prune3
        if s == kSett.tok: continue                                              # prune3
        ans.append([s, prune3(children, depth+1), num, depth])                   # prune3
    return ans                                                                   # prune3
def prune(x): return prune3(prune2(prune1(x))[0])                                # prune
def prune_makeup(x):                                                             # prune_makeup
    for elem in x:                                                               # prune_makeup
        yield [elem[0], prune_makeup(elem[1]), 0] if len(elem) > 1 else [elem[0], [], 0] # last column for extra metadata slot # prune_makeup
class create(BaseCli):                                                           # create
    def __init__(self):                                                          # create
        """Creates a tree structure from a flat table.
Example::

    flat = [["a", "b"],
            ["a", "b", "2"],
            ["a", "b", "3"],
            ["a", "c", "4"],
            ["a", "c", "5"],
            ["a", "d"]]
    flat | ktree.create()

That will return this::

    [['a',
      [['b', [
         ['2', [], 1, 2],
         ['3', [], 1, 2]],
       1, 1],
       ['c', [
         ['4', [], 1, 2],
         ['5', [], 1, 2]],
       0, 1],
       ['d', [], 1, 1]],
     0, 0]]

Each element is represented by a 3-tuple: ``[name, children, ends here?, depth]``.
Children is just a list of elements, and "ends here?" means whether there
is a row in the flat table that exists for this particular element. Most
has it, like "a/b/2" and "a/b/3", but some don't, like "a/c", so "ends here?"
of "c" is 0.

Everything in the :mod:`ktree` module is relatively unoptimized, because my use
cases don't really require super high speed, so right now it can only handle
18k rows/second.

.. warning::

    Internally, this makes use of the special character "\\ue000", configurable
    at ``settings.cli.ktree.tok`` to create the tree. This is in Unicode's private
    use area, so it shouldn't appear anywhere in normal text. However, if your
    application uses this character, it's best to change it to something else
"""                                                                              # create
    def __ror__(self, it): return prune(it | cli.transpose(fill=kSett.tok) | cli.transpose() | mktree1()) # create
def _applyRet(x, f) -> "ch":                                                     # _applyRet
    ans = []                                                                     # _applyRet
    for s, ch, *other in x: ans.append(f(s, _applyRet(ch, f), *other))           # _applyRet
    return ans                                                                   # _applyRet
class apply(BaseCli):                                                            # apply
    def __init__(self, f):                                                       # apply
        """Transforms the tree recursively.
Example::

    a = flat | ktree.create() | ktree.apply(lambda s,ch,e,d: [s,ch,(s | (tryout(0) | aS(int))) + (ch | cut(2) | toSum())])

That returns this::

    [['a',
      [['b', [['2', [], 2], ['3', [], 3]], 5],
       ['c', [['4', [], 4], ['5', [], 5]], 9],
       ['d', [], 0]],
      14]]

Extracting the final result::

    a | cut(2) | item() # returns 14

Here, I'm trying to add up all elements that can be converted into an
integer, but do that at every node in the tree. The ``s | (tryout(0) | aS(int))``
part is trying to convert it into an integer, returning 0 if it can't do it.
Then, the ``ch | cut(2) | toSum()`` tries to add up the result of all children
underneath it. The results naturally trickle upwards and you get the result "14"
at the top level, at which point it's trivial to extract.

If you're using a lot of cli tools inside the lambda function and the performance
is terrible, you can try to initialize the cli tools outside the lambda, and then
just use it inside, to eliminate cli initialization cost, like this::

    def getVal(x):
        try: return int(x)
        except: return 0
    sumF = cut(2) | toSum()
    flat | ktree.create() | ktree.apply(lambda s,ch,e,d: [s,ch,getVal(s) + sumF(ch)])

This should give you enough performance for your purpose, but if you want even more
perf, don't use cli tools inside the lambda at all."""                           # apply
        self.f = f                                                               # apply
    def __ror__(self, it): return _applyRet(it, self.f)                          # apply
def _filter(x, f) -> "ch":                                                       # _filter
    ans = []                                                                     # _filter
    for s, ch, *other in x:                                                      # _filter
        ch = _filter(ch, f)                                                      # _filter
        res = f(s, ch, *other)                                                   # _filter
        if res: ans.append([s, ch, *other])                                      # _filter
    return ans                                                                   # _filter
class filt(BaseCli):                                                             # filt
    def __init__(self, f):                                                       # filt
        """Filters out elements that don't pass a certain predicate.
Example::

    # returns [['a', [['b', [], 1, 1], ['c', [], 0, 1], ['d', [], 1, 1]], 0, 0]]
    flat | ktree.create() | ktree.filt(lambda s,ch,e,d: d<2)

Essentially, this filters out any elements that are of depth 2 or more."""       # filt
        self.f = f                                                               # filt
    def __ror__(self, it): return _filter(it, self.f)                            # filt
class select(BaseCli):                                                           # select
    def __init__(self, term:str, depth:int, exact:bool=True):                    # select
        """Selects for a single element at a particular depth.
Example::

    # returns [['a', [['b', [['2', [], 1, 2], ['3', [], 1, 2]], 1, 1]], 0, 0]]. It selects for "/*/b"
    flat | ktree.create() | ktree.select("b", 1)
    # returns [['a', [['b', [], 1, 1], ['c', [], 0, 1], ['d', [], 1, 1]], 0, 0]]. It selects for "/*/*/b", which does not exist. But it still returns all elements in depth 1, as depth 1 is not selected for
    flat | ktree.create() | ktree.select("b", 2)
    # returns [['a', [['b', [['2', [], 1, 2], ['3', [], 1, 2]], 1, 1]], 0, 0]]. It selects for "/a/b"
    flat | ktree.create() | ktree.select("a", 0) | ktree.select("b", 1)
    # returns []. It selects for "/c/b", which does not exist
    flat | ktree.create() | ktree.select("c", 0) | ktree.select("b", 1)

:param term: term to select
:param depth: depth at which to select term
:param exact: if True, select for term exactly, else select for elements that contains term""" # select
        self.term = term; self.depth = depth; self.exact = exact                 # select
    def __ror__(self, it):                                                       # select
        term = self.term; depth = self.depth; exact = self.exact                 # select
        if exact: return it | filt(lambda s,ch,e,d: (d != depth) or (d == depth and term == s)) # select
        else:     return it | filt(lambda s,ch,e,d: (d != depth) or (d == depth and term in s)) # select
def _flatten(x, joinF, prefix=""):                                               # _flatten
    for s,ch,*other in x: pre = joinF(prefix, s); yield [pre,*other]; yield from _flatten(ch, joinF, prefix=pre) # _flatten
class flatten(BaseCli):                                                          # flatten
    def __init__(self, joinF=None):                                              # flatten
        """Flattens the tree.
Example::

    flat | ktree.create() | ktree.flatten() | deref()

This returns::

    [['/a', 0, 0],
     ['/a/b', 1, 1],
     ['/a/b/2', 1, 2],
     ['/a/b/3', 1, 2],
     ['/a/c', 0, 1],
     ['/a/c/4', 1, 2],
     ['/a/c/5', 1, 2],
     ['/a/d', 1, 1]]

:param joinF: a function that takes in 2 parameters: prefix (like "/a/b") and element
    name (like "2"), to form a new prefix ("/a/b/2"). Defaulted to ``lambda p,s: f"{p}/{s}"``
"""                                                                              # flatten
        self.joinF = joinF or (lambda p,s: f"{p}/{s}")                           # flatten
    def __ror__(self, it): return _flatten(it, self.joinF)                       # flatten
def _prettyDepth(x, depth=0):                                                    # _prettyDepth
    for s,ch,*other in x:                                                        # _prettyDepth
        yield f"{depth}".ljust(3) + f"  "*(depth+1) + ([s, *other] | cli.join(' - ')) # _prettyDepth
        yield from _prettyDepth(ch, depth+1)                                     # _prettyDepth
def _prettyNoDepth(x, depth=0):                                                  # _prettyNoDepth
    for s,ch,*other in x:                                                        # _prettyNoDepth
        yield f"  "*depth + f"{s} - {other | cli.join(' - ')}"                   # _prettyNoDepth
        yield from _prettyNoDepth(ch, depth+1)                                   # _prettyNoDepth
class pretty(BaseCli):                                                           # pretty
    def __init__(self, depth=True):                                              # pretty
        """Converts tree into a string iterator, indented
with the correct level of depth. See also: :class:`display`

:param depth: whether to display the depth at the start of each line or not"""   # pretty
        self.depth = depth                                                       # pretty
    def __ror__(self, it): return _prettyDepth(it) if self.depth else _prettyNoDepth(it) # pretty
class display(BaseCli):                                                          # display
    def __init__(self, lines=10, depth=True):                                    # display
        """Displays the tree.
Example::

    flat | ktree.create() | ktree.display()
    # equivalent to the above
    flat | ktree.create() | ktree.pretty() | stdout()

:param lines: line limit. Put "None" to display all
:param depth: whether to display the depth at the start of each line or not"""   # display
        self.lines = lines; self.depth = depth                                   # display
    def __ror__(self, it): it | pretty(self.depth) | cli.head(self.lines) | cli.stdout() # display
def search(term:str, exact=False):                                               # search
    """Searches for a specific term in the name of all elements.
Example::

    # returns [['a', [['b', [['2', [], 1, 2]], 1, 1]], 0, 0]]
    flat | ktree.create() | ktree.search("2")

:param term: term to search for
:param exact: if True, searches for that exact term, else searches
    for the term inside the string"""                                            # search
    if exact: st1 = apply(lambda s,ch,*o: [s,ch,(term == s) or any(c[2] for c in ch),*o]) # search
    else:     st1 = apply(lambda s,ch,*o: [s,ch,(term in s) or any(c[2] for c in ch),*o]) # search
    return st1 | filt(lambda s,ch,e,*o: e) | apply(lambda s,ch,e,*o: [s,ch,*o])  # search
_walk = lambda f: cli.aS(lambda fn: os.walk(os.path.expanduser(fn)) | cli.cut(0, 2) | cli.ungroup() | cli.join(os.sep).all() | f | cli.op().split(os.sep).all() | create()) # search
def walk(fn=None, f=None):                                                       # walk
    """Walks the file system and generates a tree of all files and folders.
Example::

    "/home/ubuntu" | ktree.walk() # returns a tree
    ktree.walk("/home/ubundu")    # same as above
    ktree.walk(".")               # can also do relative paths
    "/home/ubuntu" | ktree.walk(f=~grep("git")) # returns a tree with no hidden git-related files
    ktree.walk("/home/ubuntu", ~grep("git"))    # same as above

:param fn: folder name
:param f: optional function if you want to filter out paths (Iterator[str])"""   # walk
    f = f or cli.iden(); return _walk(f) if fn is None else fn | _walk(f)        # walk



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import k1lib; from k1lib.cli import empty, shape
from .callbacks import Callback, Callbacks, Cbs
from typing import Tuple, List
try: import torch; hasTorch = True
except: torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["Recorder"]
@k1lib.patch(Cbs)
class Recorder(Callback):                                                        # Recorder
    """Records xb, yb and y from a short run. No training involved.
Example::

    l = k1lib.Learner.sample()
    l.cbs.add(Cbs.Recorder())
    xbs, ybs, ys = l.Recorder.record(1, 2)
    xbs # list of x batches passed in
    ybs # list of y batches passed in, "the correct label"
    ys #  list of network's output

If you have extra metadata in your dataloader, then the recorder will return
(xb, yb, metab, ys) instead::

    # creating a new dataloader that yields (xb, yb, metadata)
    x = torch.linspace(-5, 5, 1000); meta = torch.tensor(range(1000))
    dl = [x, x+2, meta] | transpose() | randomize(None) | repeatFrom() | batched()\
    | (transpose() | (toTensor() + toTensor() + toTensor())).all() | stagger(50)

    l = k1lib.Learner.sample(); l.data = [dl, []]
    l.cbs.add(Cbs.Recorder())
    xbs, ybs, metabs, ys = l.Recorder.record(1, 2)
"""                                                                              # Recorder
    def __init__(self):                                                          # Recorder
        super().__init__(); self.order = 20; self.suspended = True               # Recorder
    def startRun(self):                                                          # Recorder
        self.xbs = []; self.ybs = []; self.metabs = []; self.ys = []             # Recorder
    def startBatch(self):                                                        # Recorder
        self.xbs.append(self.l.xb.detach())                                      # Recorder
        self.ybs.append(self.l.yb.detach())                                      # Recorder
        self.metabs.append(self.l.metab)                                         # Recorder
    def endRun(self):                                                            # Recorder
        n = min(len(self.xbs), len(self.ybs), len(self.metabs), len(self.ys))    # Recorder
        self.xbs = self.xbs[:n]; self.ybs = self.ybs[:n]                         # Recorder
        self.metabs = self.metabs[:n]; self.ys = self.ys[:n]                     # Recorder
    def endPass(self):                                                           # Recorder
        self.ys.append(self.l.y.detach())                                        # Recorder
    @property                                                                    # Recorder
    def values(self):                                                            # Recorder
        hasMeta = self.metabs | ~empty() | shape(0) > 0                          # Recorder
        if hasMeta: return self.xbs, self.ybs, self.metabs, self.ys              # Recorder
        else: return self.xbs, self.ybs, self.ys                                 # Recorder
    def record(self, epochs:int=1, batches:int=None) -> Tuple[List[torch.Tensor], List[torch.Tensor], List[torch.Tensor]]: # Recorder
        """Returns recorded xBatch, yBatch and answer y"""                       # Recorder
        self.suspended = False                                                   # Recorder
        try:                                                                     # Recorder
            with self.cbs.context(), self.cbs.suspendEval():                     # Recorder
                self.cbs.add(Cbs.DontTrain()).add(Cbs.TimeLimit(5))              # Recorder
                self.l.run(epochs, batches)                                      # Recorder
        finally: self.suspended = True                                           # Recorder
        return self.values                                                       # Recorder
    def __repr__(self):                                                          # Recorder
        return f"""{self._reprHead}, can...
- r.record(epoch[, batches]): runs for a while, and records x and y batches, and the output
{self._reprCan}"""                                                               # Recorder



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""These are pretty standard imports that I always
do. I put it here so that I can do::

    from k1lib.imports import *

and everything will be in place"""
import k1lib; settings = k1lib.settings; cli = k1lib.cli
import numpy as np
if settings.startup.import_optionals:
    try:
        import torch; from torch import nn, optim
        import torch.nn.functional as F, torch.utils.data as data
    except: pass
    try:
        import matplotlib.pyplot as plt, matplotlib as mpl; import matplotlib.cm as cm
        plt.rcParams['figure.dpi'] = 100; plt.rcParams["animation.html"] = "jshtml"
    except: pass
    import dill as pickle, multiprocessing as mp, concurrent.futures as futures, threading
    import math, os, time, sys, random, logging, traceback, re, typing, glob, warnings, asyncio, ast
    import dill, json, inspect, xml, base64, io, html
    import functools; from functools import partial, lru_cache
    import contextlib; from contextlib import contextmanager
    from collections import deque, defaultdict
    from typing import List, Tuple, Callable, Union, Iterator, Set, Dict, Any
    from k1lib import schedule, graphEqn, mo, knn, fmt, selector, viz, Cbs, _k1a, serve, p5, k1ui, selen, kws, kast, zircon, serpent
    k1a = _k1a
    for e in cli._scatteredClis: globals()[e.__name__] = e
    if "py_k1lib_in_applyMp" not in os.environ: k1lib.dontWrap()
from k1lib.cli import *; k1 = k1lib
from math import e, pi; inf = float("inf"); nan = float("nan"); # this section is for constants
h = 6.62607015e-34; hbar = h/(2*pi); Na = 6.0221408e23; kb = 1.380649e-23
c = 299_792_458; qe = 1.60217663e-19; me = 9.1093837e-31; mn = 1.67262192e-27
e0 = 8.85418782e-12; Dal = u = 1.6605390666e-27; R = 8.3145; sb = 5.670374e-8
secondsInYear = 31_556_926; pm = "±"
if settings.startup.or_patch.numpy: cli.init.patchNumpy()
if settings.startup.or_patch.dict: cli.init.patchDict()
if settings.startup.import_optionals:
    try:
        import pandas as pd
        if settings.startup.or_patch.pandas: cli.init.patchPandas()
    except ImportError as _: pass
    try:
        import ray
        if settings.startup.init_ray: ray.init()
    except: pass
def dummy():                                                                     # dummy
    """Does nothing. Only here so that you can read source code of this file
and see what's up."""                                                            # dummy
    pass                                                                         # dummy
if settings.startup.import_optionals:                                            # dummy
    try: # these are optional imports                                            # dummy
        import torchvision as vision                                             # dummy
        import torchvision.datasets as tvDs                                      # dummy
        import torchvision.transforms as tf                                      # dummy
    except: pass                                                                 # dummy
    try: import datetime                                                         # dummy
    except: pass                                                                 # dummy
    try: import IPython                                                          # dummy
    except: pass                                                                 # dummy
    try: import PIL                                                              # dummy
    except: pass                                                                 # dummy
    try: from scipy.optimize import fsolve; import scipy                         # dummy
    except: pass                                                                 # dummy
    try: import einops                                                           # dummy
    except: pass                                                                 # dummy
    try: from tqdm import tqdm                                                   # dummy
    except: pass                                                                 # dummy
    try: import sympy                                                            # dummy
    except: pass                                                                 # dummy
    try: import rdkit; from rdkit import Chem                                    # dummy
    except: pass                                                                 # dummy
    try: import requests                                                         # dummy
    except: pass                                                                 # dummy
    try: import seaborn as sns                                                   # dummy
    except: pass                                                                 # dummy
    try: import pyarrow; import pyarrow.feather as feather                       # dummy
    except: pass                                                                 # dummy
    try: import plotly; import plotly.express as px                              # dummy
    except: pass                                                                 # dummy
    try: import graphviz                                                         # dummy
    except: pass                                                                 # dummy
    try: import line_profiler; line_profiler.load_ipython_extension(IPython.get_ipython()) # dummy
    except: pass                                                                 # dummy
    try: import cython; cython.load_ipython_extension(IPython.get_ipython())     # dummy
    except: pass                                                                 # dummy
    try: import websockets                                                       # dummy
    except: pass                                                                 # dummy
    try: import bs4                                                              # dummy
    except: pass                                                                 # dummy
    try: import cv2; import cv2.aruco as aruco                                   # dummy
    except: pass                                                                 # dummy
    try: import cloudpickle                                                      # dummy
    except: pass                                                                 # dummy



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import k1lib, dill, traceback
from k1lib.callbacks import Cbs
from typing import Union
from time import time as _time
try: import torch; import torch.nn as nn; hasTorch = True
except:
    torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {}))
    nn = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
__all__ = ["CancelRunException", "CancelEpochException", "CancelBatchException",
           "Learner"]
class CancelRunException(Exception):                                             # CancelRunException
    """Used in core training loop, to skip the run entirely"""                   # CancelRunException
    pass                                                                         # CancelRunException
class CancelEpochException(Exception):                                           # CancelEpochException
    """Used in core training loop, to skip to next epoch"""                      # CancelEpochException
    pass                                                                         # CancelEpochException
class CancelBatchException(Exception):                                           # CancelBatchException
    """Used in core training loop, to skip to next batch"""                      # CancelBatchException
    pass                                                                         # CancelBatchException
def _tab(text:Union[list, str], pad="    ") -> Union[list, str]:                 # _tab
    if isinstance(text, str): # this is old function that's replaced in main lib, but still useful # _tab
        return "\n".join([pad + line for line in text.split("\n")])              # _tab
    else: return [pad + line for line in text]                                   # _tab
class Learner:                                                                   # Learner
    def __init__(self):                                                          # Learner
        self._model = None; self._data = None; self._opt = None                  # Learner
        self._cbs = None; self.fileName = None                                   # Learner
        self.css = "*"; self.exceptionRaised = None # slowly pops                # Learner
        self.cbs = k1lib.Callbacks().withBasics().withQOL().withAdvanced()       # Learner
    @property                                                                    # Learner
    def model(self):                                                             # Learner
        """Set this to change the model to run"""                                # Learner
        return self._model                                                       # Learner
    @model.setter                                                                # Learner
    def model(self, model): self._model = model                                  # Learner
    @property                                                                    # Learner
    def data(self):                                                              # Learner
        """Set this to change the data (list of 2 dataloader) to run against.""" # Learner
        return self._data                                                        # Learner
    @data.setter                                                                 # Learner
    def data(self, data): self._data = data                                      # Learner
    @property                                                                    # Learner
    def opt(self):                                                               # Learner
        """Set this to change the optimizer. If you're making your own
optimizers, beware to follow the PyTorch's style guide as there are callbacks
that modifies optimizer internals while training like
:class:`k1lib.schedule.ParamScheduler`."""                                       # Learner
        return self._opt                                                         # Learner
    @opt.setter                                                                  # Learner
    def opt(self, opt): self._opt = opt                                          # Learner
    @property                                                                    # Learner
    def cbs(self):                                                               # Learner
        """The :class:`~k1lib.callbacks.callbacks.Callbacks` object. Initialized to
include all the common callbacks. You can set a new one if you want to."""       # Learner
        return self._cbs                                                         # Learner
    @cbs.setter                                                                  # Learner
    def cbs(self, cbs): cbs.l = self; self._cbs = cbs                            # Learner
    @property                                                                    # Learner
    def css(self) -> str:                                                        # Learner
        """The css selector string. Set this to select other parts of the network.
After setting, you can access the selector like this: :code:`l.selector`

See also: :class:`~k1lib.selector.ModuleSelector`"""                             # Learner
        return self._css                                                         # Learner
    @css.setter                                                                  # Learner
    def css(self, css:str):                                                      # Learner
        self._css = css                                                          # Learner
        if self.model != None: self.selector = k1lib.selector.select(self.model, self.css) # Learner
    @property                                                                    # Learner
    def lossF(self):                                                             # Learner
        """Set this to specify a loss function."""                               # Learner
        raise NotImplementedError("lossF actually doesn't really exist. Used to exist as a core part of Learner, but then has been converted to Cbs.LossF") # Learner
    @lossF.setter                                                                # Learner
    def lossF(self, lossF):                                                      # Learner
        if hasattr(self.cbs, "LossF"): self.cbs.LossF.lossF = lossF              # Learner
        else: self.cbs.add(Cbs.LossF(lossF))                                     # Learner
    def __getattr__(self, attr):                                                 # Learner
        if attr == "cbs": raise AttributeError()                                 # Learner
        return getattr(self.cbs, attr)                                           # Learner
    def __getstate__(self):                                                      # Learner
        answer = dict(self.__dict__); answer.pop("selector", None)               # Learner
        answer.pop("_data", None); return answer                                 # Learner
    def __setstate__(self, state):                                               # Learner
        self.__dict__.update(state)                                              # Learner
        self.__dict__["_data"] = None                                            # Learner
        self.css = self.css; self.cbs.l = self                                   # Learner
    def evaluate(self): pass # supposed to be overriden, to provide functionality here # Learner
    @property                                                                    # Learner
    def _warnings(self):                                                         # Learner
        warnings = "Warning: no model yet. Set using `l.model = ...`\n" if self.model == None else "" # Learner
        lossClasses = tuple([*k1lib.Callback.lossCls])                           # Learner
        lossFnCbs = [True for cb in self.cbs if isinstance(cb, lossClasses)]     # Learner
        warnings += "Warning: no loss function callback detected (or you set `lossF` already but then erased all callbacks)! Set using `l.lossF = ...` or `l.cbs.add(Cbs.LossF(...))`\n" if len(lossFnCbs) == 0 else "" # Learner
        warnings += "Warning: no data yet. Set using `l.data = ...`\n" if self.data == None else "" # Learner
        warnings += "Warning: no optimizer yet. Set using `l.opt = ...`\n" if self.opt == None else "" # Learner
        if warnings != "": warnings += "\n\n"                                    # Learner
        return warnings                                                          # Learner
    def __dir__(self):                                                           # Learner
        answer = list(super().__dir__())                                         # Learner
        answer.extend(self.cbs.cbsDict.keys()); return answer                    # Learner
    def __repr__(self):                                                          # Learner
        return f"""{self._warnings}l.model:\n{_tab(k1lib.limitLines(str(self.model)))}
l.opt:\n{_tab(k1lib.limitLines(str(self.opt)))}
l.cbs:\n{_tab(k1lib.limitLines(self.cbs.__repr__()))}
Use...
- l.model = ...: to specify a nn.Module object
- l.data = ...: to specify data object
- l.opt = ...: to specify an optimizer
- l.lossF = ...: to specify a loss function
- l.css = ...: to select modules using CSS. "#root" for root model
- l.cbs = ...: to use a custom `Callbacks` object
- l.selector: to get the modules selected by `l.css`
- l.run(epochs): to run the network
- l.Loss: to get a specific callback, this case "Loss"\n\n"""                    # Learner
@k1lib.patch(Learner)                                                            # Learner
def save(self, fileName:str=None):                                               # save
    """Saves this :class:`Learner` to file. See also: :meth:`load`. Does not
save the ``data`` object, because that's potentially very big. Example::

    l = k1.Learner()
    # saves learner to "skip1_128bs.pth" and model to "skip1_128bs.model.pth"
    l.save("skip1_128bs")

:param fileName: name to save file into"""                                       # save
    torch.save(self, f"{fileName}.pth", pickle_module=dill)                      # save
    torch.save(self.model, f"{fileName}.model.pth", pickle_module=dill)          # save
    print(f"Saved to {fileName}")                                                # save
@k1lib.patch(Learner, static=True)                                               # save
def load(fileName:str=None):                                                     # load
    """Loads a :class:`Learner` from a file. See also: :meth:`save`.
Example::

    # this will load up learner in file "skip1_128bs.pth"
    l = k1.Learner.load("skip1_128bs")

:param fileName: if empty, then will prompt for file name"""                     # load
    f = fileName or input("Enter learner file name to load:")                    # load
    print(f"Loaded from {f}"); return torch.load(f"{f}.pth", pickle_module=dill) # load
@k1lib.patch(Learner)                                                            # load
def _run1Batch(self):                                                            # _run1Batch
    self.cbs("startBatch")                                                       # _run1Batch
    try:                                                                         # _run1Batch
        self.cbs("startPass", "inPass", "endPass")                               # _run1Batch
        self.cbs("startLoss", "inLoss", "endLoss")                               # _run1Batch
        if not self.cbs("startBackward"): self.lossG.backward()                  # _run1Batch
        if not self.cbs("startStep"):  self.opt.step()                           # _run1Batch
        if not self.cbs("startZeroGrad"): self.opt.zero_grad(set_to_none=True)   # _run1Batch
    except k1lib.CancelBatchException as ex:                                     # _run1Batch
        self.cbs("cancelBatch"); print(f"Batch cancelled: {ex}.", end="\n" if k1lib.settings.cancelRun_newLine else "") # _run1Batch
    except (k1lib.CancelEpochException, k1lib.CancelRunException) as ex:         # _run1Batch
        # makes sure cancelBatch and endBatch gets called, for potential         # _run1Batch
        # cleanups, then reraise the exception                                   # _run1Batch
        self.cbs("cancelBatch", "endBatch"); raise ex                            # _run1Batch
    self.cbs("endBatch")                                                         # _run1Batch
class DI: # data interceptor, just to record data loading times                  # DI
    def __init__(self, l:Learner, data): self.l = l; self.data = data            # DI
    def __len__(self): return len(self.data)                                     # DI
    def __iter__(self):                                                          # DI
        try:                                                                     # DI
            data = iter(self.data); timings = self.l.cbs.timings                 # DI
            while True:                                                          # DI
                beginTime = _time(); d = next(data)                              # DI
                timings.loadData += _time() - beginTime; yield d                 # DI
        except StopIteration: pass                                               # DI
@k1lib.patch(Learner)                                                            # DI
def _run1Epoch(self):                                                            # _run1Epoch
    self.cbs("startEpoch")                                                       # _run1Epoch
    try:                                                                         # _run1Epoch
        train, valid = self.data; train = DI(self, train); valid = DI(self, valid) # _run1Epoch
        try: self.batches = len(train) + len(valid)                              # _run1Epoch
        except: pass                                                             # _run1Epoch
        self.model.train()                                                       # _run1Epoch
        for self.batch, (self.xb, self.yb, *self.metab) in enumerate(train):     # _run1Epoch
            self._run1Batch()                                                    # _run1Epoch
        trainLen = self.batch + 1                                                # _run1Epoch
        if not self.cbs("startValidBatches"):                                    # _run1Epoch
            self.model.eval();                                                   # _run1Epoch
            for self.batch, (self.xb, self.yb, *self.metab) in enumerate(valid): # _run1Epoch
                self.batch += trainLen; self._run1Batch()                        # _run1Epoch
        if self.batches is None: self.batches = self.batch + 1                   # _run1Epoch
    except k1lib.CancelEpochException as ex:                                     # _run1Epoch
        self.cbs("cancelEpoch"); print(f"Epoch cancelled: {ex}.", end="\n" if k1lib.settings.cancelRun_newLine else "") # _run1Epoch
    except k1lib.CancelRunException as ex:                                       # _run1Epoch
        self.cbs("cancelEpoch", "endEpoch"); raise ex                            # _run1Epoch
    self.cbs("endEpoch")                                                         # _run1Epoch
@k1lib.patch(Learner)                                                            # _run1Epoch
def run(self, epochs:int, batches:int=None):                                     # run
    """Main run function.

:param epochs: number of epochs to run. 1 epoch is the length of the dataset
:param batches: if set, then cancels the epoch after reaching the specified batch""" # run
    if self._warnings != "":                                                     # run
        if not input(f"""You still have these warnings:\n\n{self._warnings}
Do you want to continue? (y/n) """).lower().startswith("y"):                     # run
            print("Run ended"); return                                           # run
    self.epochs = int(epochs); self.batches = None                               # run
    self.css = self.css # update module selector                                 # run
    with self.cbs.context():                                                     # run
        if batches is not None: self.cbs.add(Cbs.BatchLimit(int(batches)))       # run
        self.cbs("startRun")                                                     # run
        try:                                                                     # run
            for self.epoch in range(self.epochs): self._run1Epoch()              # run
        except k1lib.CancelRunException as ex:                                   # run
            self.cbs("cancelRun"); print(f"Run cancelled: {ex}.", end="\n" if k1lib.settings.cancelRun_newLine else "") # run
        self.cbs("endRun"); return self                                          # run
@k1lib.patch(Learner)                                                            # run
def __call__(self, xb, yb=None):                                                 # __call__
    """Executes just a small batch. Convenience method to query how the network is
doing.

:param xb: x batch
:param yb: y batch. If specified, return (y, loss), else return y alone
"""                                                                              # __call__
    oldData = self.data; self.data = [[(xb, (yb or torch.tensor(0)))], []]       # __call__
    with self.cbs.suspendEval(), self.cbs.context():                             # __call__
        ex = lambda _: k1lib.raiseEx(k1lib.CancelBatchException)                 # __call__
        self.cbs.add(k1lib.Callback().withCheckpoint("startLoss" if yb is None else "startBackward", ex)) # __call__
        self.run(1, 1)                                                           # __call__
    self.data = oldData; return self.y if yb is None else (self.y, self.loss)    # __call__
@k1lib.patch(Learner)                                                            # __call__
def evaluate(self):                                                              # evaluate
    """Function to visualize quickly how the network is doing. Undefined by default,
just placed here as a convention, so you have to do something like this::

    l = k1lib.Learner()
    def evaluate(self):
        xbs, ybs, ys = self.Recorder.record(1, 3)
        plt.plot(torch.vstack(xbs), torch.vstack(ys))
    l.evaluate = partial(evaluate(l))
"""                                                                              # evaluate
    raise NotImplementedError("You have to define evaluate() by yourself")       # evaluate
from k1lib.cli import *                                                          # evaluate
@k1lib.patch(Learner, static=True)                                               # evaluate
def sample() -> Learner:                                                         # sample
    """Creates an example learner, just for simple testing stuff anywhere. The
network tries to learn the function y=x. Only bare minimum callbacks are included.""" # sample
    l = Learner(); x = torch.linspace(-5, 5, 1000)                               # sample
    l.data = [x, x] | transpose() | randomize(None) | splitW() | (repeatFrom() | randomize() | batched(32) | (transpose() | toTensor()).all()).all() | stagger.tv(300) | toList() # sample
    class Model(torch.nn.Module):                                                # sample
        def __init__(self):                                                      # sample
            super().__init__()                                                   # sample
            self.lin1 = k1lib.knn.LinBlock(1, 3)                                 # sample
            self.lin2 = nn.Linear(3, 1)                                          # sample
        def forward(self, x):                                                    # sample
            return ((x[:, None] + 2) | self.lin1 | self.lin2).squeeze()          # sample
    l.model = Model(); l.cbs = k1lib.Callbacks().add(Cbs.CoreNormal()).add(Cbs.Loss()).add(Cbs.ProgressBar()) # sample
    l.lossF = lambda y, yb: ((y - yb) ** 2).sum()                                # sample
    l.opt = torch.optim.Adam(l.model.parameters(), lr=3e-3); return l            # sample



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
"""
This module is for nice visualization tools. This is exposed automatically with::

   from k1lib.imports import *
   viz.mask # exposed
"""
import k1lib, base64, io, os, matplotlib as mpl, warnings
import k1lib.cli as cli
plt = k1lib.dep("matplotlib.pyplot"); import numpy as np
from typing import Callable, List, Union
from functools import partial, update_wrapper
try: import torch; import torch.nn as nn; hasTorch = True
except:
    torch = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {}))
    nn = k1lib.Object().withAutoDeclare(lambda: type("RandomClass", (object, ), {})); hasTorch = False
try: import PIL; hasPIL = True
except: hasPIL = False
__all__ = ["SliceablePlot", "plotSegments", "Carousel", "Toggle", "ToggleImage",
           "Scroll", "confusionMatrix", "FAnim", "mask", "PDF"]
class _PlotDecorator:                                                            # _PlotDecorator
    """The idea with decorators is that you can do something like this::

    sp = k1lib.viz.SliceablePlot()
    sp.yscale("log") # will format every plot as if ``plt.yscale("log")`` has been called

This class is not expected to be used by end users though."""                    # _PlotDecorator
    def __init__(self, sliceablePlot:"SliceablePlot", name:str):                 # _PlotDecorator
        """
:param sliceablePlot: the parent plot
:param name: the decorator's name, like "yscale" """                             # _PlotDecorator
        self.sliceablePlot = sliceablePlot                                       # _PlotDecorator
        self.name = name; self.args, self.kwargs = None, None                    # _PlotDecorator
    def __call__(self, *args, **kwargs):                                         # _PlotDecorator
        """Stores all args, then return the parent :class:`SliceablePlot`"""     # _PlotDecorator
        self.args = args; self.kwargs = kwargs; return self.sliceablePlot        # _PlotDecorator
    def run(self): getattr(plt, self.name)(*self.args, **self.kwargs)            # _PlotDecorator
class SliceablePlot:                                                             # SliceablePlot
    """This is a plot that is "sliceable", meaning you can focus into
a particular region of the plot quickly. A minimal example looks something
like this::

    import numpy as np, matplotlib.pyplot as plt, k1lib
    x = np.linspace(-2, 2, 100)

    def normalF():
        plt.plot(x, x**2)

    @k1lib.viz.SliceablePlot.decorate
    def plotF(_slice):
        plt.plot(x[_slice], (x**2)[_slice])

    plotF()[70:] # plots x^2 equation with x in [0.8, 2]

So, ``normalF`` plots the equation :math:`x^2` with x going from -2 to 2.
You can convert this into a :class:`SliceablePlot` by adding a term of
type :class:`slice` to the args, and decorate with :meth:`decorate`. Now,
every time you slice the :class:`SliceablePlot` with a specific range,
``plotF`` will receive it.

How intuitive everything is depends on how you slice your data. ``[70:]``
results in x in [0.8, 2] is rather unintuitive. You can change it into
something like this::

    @k1lib.viz.SliceablePlot.decorate
    def niceF(_slice):
        n = 100; r = k1lib.Range(-2, 2)
        x = np.linspace(*r, n)
        _slice = r.toRange(k1lib.Range(n), r.bound(_slice)).slice_
        plt.plot(x[_slice], (x**2)[_slice])
    # this works without a decorator too btw: k1lib.viz.SliceablePlot(niceF)

    niceF()[0.3:0.7] # plots x^2 equation with x in [0.3, 0.7]
    niceF()[0.3:] # plots x^2 equation with x in [0.3, 2]

The idea is to just take the input :class:`slice`, put some bounds on
its parts, then convert that slice from [-2, 2] to [0, 100]. Check
out :class:`k1lib.Range` if it's not obvious how this works.

A really cool feature of :class:`SliceablePlot` looks like this::

    niceF().legend(["A"])[-1:].grid(True).yscale("log")

This will plot :math:`x^2` with range in [-1, 2] with a nice grid, and
with y axis's scale set to log. Essentially, undefined method calls
on a :class:`SliceablePlot` will translate into ``plt`` calls. So the
above is roughly equivalent to this::

    x = np.linspace(-2, 2, 100)
    plt.plot(x, x**2)
    plt.legend(["A"])
    plt.grid(True)
    plt.yscale("log")

.. image:: images/SliceablePlot.png

This works even if you have multiple axes inside your figure. It's
wonderful, isn't it?"""                                                          # SliceablePlot
    def __init__(self, plotF:Callable[[slice], None], slices:Union[slice, List[slice]]=slice(None), plotDecorators:List[_PlotDecorator]=[], docs=""): # SliceablePlot
        """Creates a new SliceablePlot. Only use params listed below:

:param plotF: function that takes in a :class:`slice` or tuple of :class:`slice`s
:param docs: optional docs for the function that will be displayed in :meth:`__repr__`""" # SliceablePlot
        self.plotF = plotF                                                       # SliceablePlot
        self.slices = [slices] if isinstance(slices, slice) else slices          # SliceablePlot
        self.docs = docs; self.plotDecorators = list(plotDecorators)             # SliceablePlot
    @staticmethod                                                                # SliceablePlot
    def decorate(f):                                                             # SliceablePlot
        """Decorates a plotting function so that it becomes a
SliceablePlot."""                                                                # SliceablePlot
        answer = partial(SliceablePlot, plotF=f)                                 # SliceablePlot
        update_wrapper(answer, f)                                                # SliceablePlot
        return answer                                                            # SliceablePlot
    @property                                                                    # SliceablePlot
    def squeezedSlices(self) -> Union[List[slice], slice]:                       # SliceablePlot
        """If :attr:`slices` only has 1 element, then return that
element, else return the entire list."""                                         # SliceablePlot
        return k1lib.squeeze(self.slices)                                        # SliceablePlot
    def __getattr__(self, attr):                                                 # SliceablePlot
        if attr.startswith("_"): raise AttributeError()                          # SliceablePlot
        # automatically assume the attribute is a plt.attr method                # SliceablePlot
        dec = _PlotDecorator(self, attr)                                         # SliceablePlot
        self.plotDecorators.append(dec); return dec                              # SliceablePlot
    def __getitem__(self, idx):                                                  # SliceablePlot
        if type(idx) == slice:                                                   # SliceablePlot
            return SliceablePlot(self.plotF, [idx], self.plotDecorators, self.docs) # SliceablePlot
        if type(idx) == tuple and all([isinstance(elem, slice) for elem in idx]): # SliceablePlot
            return SliceablePlot(self.plotF, idx, self.plotDecorators, self.docs) # SliceablePlot
        raise Exception(f"Don't understand {idx}")                               # SliceablePlot
    def __repr__(self, show=True):                                               # SliceablePlot
        self.plotF(self.squeezedSlices)                                          # SliceablePlot
        for ax in plt.gcf().get_axes():                                          # SliceablePlot
            plt.sca(ax)                                                          # SliceablePlot
            for decorator in self.plotDecorators: decorator.run()                # SliceablePlot
        if show: plt.show()                                                      # SliceablePlot
        return f"""Sliceable plot. Can...
- p[a:b]: to focus on a specific range of the plot
- p.yscale("log"): to perform operation as if you're using plt{self.docs}"""     # SliceablePlot
def plotSegments(x:List[float], y:List[float], states:List[int], colors:List[str]=None): # plotSegments
    """Plots a line graph, with multiple segments with different colors.

Idea is, you have a normal line graph, but you want to color parts of
the graph red, other parts blue. Then, you can pass a "state" array, with
the same length as your data, filled with ints, like this::

    y = np.array([ 460800,  921600,  921600, 1445888, 1970176, 1970176, 2301952,
           2633728, 2633728, 3043328, 3452928, 3452928, 3457024, 3461120,
           3463680, 3463680, 3470336, 3470336, 3467776, 3869184, 3865088,
           3865088, 3046400, 2972672, 2972672, 2309632, 2504192, 2504192,
           1456128, 1393664, 1393664,  472576])
    s = np.array([1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           1, 0, 0, 1, 0, 0, 1, 0, 0, 1])
    plotSegments(None, y, s, colors=["tab:blue", "tab:red"])

.. image:: images/plotSegments.png

:param x: (nullable) list of x coordinate at each point
:param y: list of y coordinates at each point
:param states: list of color at each point
:param colors: string colors (matplotlib color strings) to display for each states""" # plotSegments
    if x is None: x = range(len(y))                                              # plotSegments
    if colors is None: colors = ["tab:blue", "tab:red", "tab:green", "tab:orange", "tab:purple", "tab:brown"][:len(x)] # plotSegments
    _x = []; _y = []; state = -1; count = -1 # stretchs, and bookkeeping nums    # plotSegments
    lx = None; ly = None # last x and y from last stretch, for plot autocompletion # plotSegments
    while count + 1 < len(x):                                                    # plotSegments
        count += 1                                                               # plotSegments
        if state != states[count]:                                               # plotSegments
            if len(_x) > 0 and state >= 0:                                       # plotSegments
                if lx != None: _x = [lx] + _x; _y = [ly] + _y                    # plotSegments
                plt.plot(_x, _y, colors[state]); lx = _x[-1]; ly = _y[-1]        # plotSegments
            _x = [x[count]]; _y = [y[count]]; state = states[count]              # plotSegments
        else: _x.append(x[count]); _y.append(y[count])                           # plotSegments
    if len(_x) > 0 and state >= 0:                                               # plotSegments
        if lx != None: _x = [lx] + _x; _y = [ly] + _y                            # plotSegments
        plt.plot(_x, _y, colors[state])                                          # plotSegments
class _Carousel:                                                                 # _Carousel
    def __init__(self, searchMode, imgs, titles):                                # _Carousel
        self.searchMode = searchMode                                             # _Carousel
        self.imgs:List[Tuple[str, str]] = imgs # Tuple[format, base64 img]       # _Carousel
        self.titles = titles                                                     # _Carousel
    def _repr_html_(self):                                                       # _Carousel
        idx = Carousel._idx(); pre = f"k1c_{idx}"; searchMode = self.searchMode  # _Carousel
        imgs = self.imgs | cli.apply(lambda x: f"`{x}`") | cli.deref(); n = len(imgs) # _Carousel
        titles = self.titles | cli.apply(lambda x: f"`{x}`") | cli.deref()       # _Carousel
        if searchMode > 0: searchBar = f"<input type='text' value='' id='{pre}_search' placeholder='Search in {'content' if searchMode == 1 else 'header'}' style='padding: 4px 4px'>" # _Carousel
        else: searchBar = ""                                                     # _Carousel
        if n > 0: contents = imgs | cli.apply(k1lib.decode) | cli.insertIdColumn() | ~cli.apply(lambda idx, html: f"<div id='{pre}_content{idx}'>{html}</div>") | cli.deref() | cli.join('\n') # _Carousel
        else: contents = "(no pages or images are found)"                        # _Carousel
        #imgs = [f"\"<img alt='' src='data:image/{fmt};base64, {img}' />\"" for fmt, img in self.imgs] # _Carousel
        html = f"""<!-- k1lib.Carousel start -->
<style>
    .{pre}_btn {{
        cursor: pointer;
        padding: 6px 12px;
        /*background: #9e9e9e;*/
        background-color: #eee;
        margin-right: 8px;
        color: #000;
        box-shadow: 0 3px 5px rgb(0,0,0,0.3);
        border-radius: 18px;
        user-select: none;
        -webkit-user-select: none; /* Safari */
        -ms-user-select: none; /* IE 10+ */
    }}
    .{pre}_btn:hover {{
        box-shadow: box-shadow: 0 3px 10px rgb(0,0,0,0.6);
        background: #4caf50;
        color: #fff;
    }}
</style>
{searchBar}
<div>
    <div style="display: flex; flex-direction: row; padding: 8px">
        <div id="{pre}_prevBtn" class="{pre}_btn">Prev</div>
        <div id="{pre}_nextBtn" class="{pre}_btn">Next</div>
    </div>
    <div id="{pre}_status" style="padding: 10px"></div>
</div>
<div id="{pre}_imgContainer">
    {contents}
</div>
<script>
    const {pre}_allImgs = [{','.join(imgs)}];
    let {pre}_imgs = [...Array({pre}_allImgs.length).keys()]; // index of all available images. If searching for something then it will be a subset of allImgs
    const {pre}_titles = [{','.join(titles)}];
    {pre}_imgIdx = 0; // n-th element of pre_imgs, not of pre_allImgs
    {pre}_searchMode = {searchMode};
    function {pre}_show(i) {{ // i here is allImgs index, not of imgs
        document.querySelector(`#{pre}_content${{i}}`).style.display = "block";
    }}
    function {pre}_hide(i) {{ // i here is allImgs index, not of imgs
        document.querySelector(`#{pre}_content${{i}}`).style.display = "none";
    }}
    function {pre}_updatePageCount() {{
        let n = {pre}_imgs.length;
        if (n > 0) document.querySelector("#{pre}_status").innerHTML = "Page: " + ({pre}_imgIdx + 1) + "/" + n;
        else document.querySelector("#{pre}_status").innerHTML = "Page: 0/0"
    }}
    function {pre}_display() {{
        let n = {pre}_imgs.length;
        for (let i = 0; i < {n}; i++) {pre}_hide(i);
        if (n > 0) {pre}_show({pre}_imgs[{pre}_imgIdx]);
        {pre}_updatePageCount();
    }};
    document.querySelector("#{pre}_prevBtn").onclick = () => {{
        {pre}_imgIdx -= 1;
        {pre}_imgIdx = Math.max({pre}_imgIdx, 0);
        {pre}_display();
    }};
    document.querySelector("#{pre}_nextBtn").onclick = () => {{
        {pre}_imgIdx += 1;
        {pre}_imgIdx = Math.min({pre}_imgIdx, {pre}_imgs.length - 1);
        {pre}_display();
    }};
    if ({pre}_searchMode > 0) {{
        {pre}_searchInp = document.querySelector("#{pre}_search");
        {pre}_searchInp.oninput = (value) => {{
            const val = {pre}_searchInp.value;
            {pre}_imgs = ({pre}_searchMode === 1 ? {pre}_allImgs : {pre}_titles).map((e, i) => [window.atob(e).includes(val), i]).filter(e => e[0]).map(e => e[1]);
            {pre}_imgIdx = 0;; {pre}_display();
        }}
    }}
    {pre}_display();
</script>
<!-- k1lib.Carousel end -->"""                                                   # _Carousel
        return html                                                              # _Carousel
class Carousel(cli.BaseCli):                                                     # Carousel
    _idx = k1lib.AutoIncrement.random()                                          # Carousel
    def __init__(self, searchMode:int=0):                                        # Carousel
        """Creates a new Carousel that can flip through a list of images/html.
Will even work even when you export the notebook as html. Example::

    x = np.linspace(-2, 2); plt.plot(x, x ** 2); im1 = plt.gcf() | toImg()
    x = np.linspace(-1, 3); plt.plot(x, x ** 2); im2 - plt.gcf() | toImg()
    im3 = "<h1>abc</h1><div>Some content</div>" # can add html
    [im1, im2, im3] | viz.Carousel() # displays in notebook cell

.. image:: images/carousel.png

There's also a builtin search functionality that works like this::

    [
        "<h1>abc</h1><div>Some content 1</div>",
        "<h1>def</h1><div>Some other content 2</div>",
        "<h1>ghi</h1><div>Another content 3</div>",
    ] | Carousel(searchMode=1)

    [
        ["<h1>abc</h1>", "<div>Some content 1</div>"],
        ["<h1>def</h1>", "<div>Some other content 2</div>"],
        ["<h1>ghi</h1>", "<div>Another content 3</div>"],
    ] | Carousel(searchMode=2)

The first mode will search for some text inside the html content. The second mode
will search inside the title only, that means it's expecting to receive Iterator[title, html/img]

:param imgs: List of initial images. Can add more images later on by using :meth:`__ror__`
:param searchMode: 0 for no search, accepts Iterator[html/img],
    1 for search content, accepts Iterator[html/img],
    2 for search title, accepts Iterator[title, html/img]
"""                                                                              # Carousel
        self.searchMode = searchMode                                             # Carousel
    def _process(self, e):                                                       # Carousel
        if isinstance(e, str): return f"{e}"                                     # Carousel
        elif hasPIL and isinstance(e, PIL.Image.Image):                          # Carousel
            return f"<img alt='' style='max-width: 100%' src='data:image/png;base64, {base64.b64encode(e | cli.toBytes()).decode()}' />" # Carousel
        else: raise Exception(f"Content is not a string nor a PIL image. Can't make a Carousel out of this unknown type: {type(e)}") # Carousel
    def __ror__(self, it):                                                       # Carousel
        imgs = []; titles = []                                                   # Carousel
        searchMode = self.searchMode                                             # Carousel
        if searchMode == 0 or searchMode == 1:                                   # Carousel
            for e in it: imgs.append(k1lib.encode(self._process(e)))             # Carousel
        elif searchMode == 2:                                                    # Carousel
            for title, e in it:                                                  # Carousel
                if not isinstance(title, str): raise Exception("Title is not a string. Can't perform search") # Carousel
                imgs.append(k1lib.encode(title+self._process(e)))                # Carousel
                titles.append(k1lib.encode(title))                               # Carousel
        else: raise Exception(f"Invalid searchMode: {searchMode}")               # Carousel
        return _Carousel(searchMode, imgs, titles)                               # Carousel
    def _jsF(self, meta):                                                        # Carousel
        if self.searchMode != 0: raise Exception("viz.Carousel._jsF() does not support .searchMode!=0. You're using the JS transpiler anyway, you can trivially build your own, more complex search engine!") # Carousel
        fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto(); imgIdx = cli.init._jsDAuto(); pre = cli.init._jsDAuto() # Carousel
        return f"""const {fIdx} = ({dataIdx}) => {{
        return unescape(`<!-- k1lib.Carousel start -->
<style>
    .{pre}_btn {{
        cursor: pointer; padding: 6px 12px; background-color: #eee;
        margin-right: 8px; color: #000;
        box-shadow: 0 3px 5px rgb(0,0,0,0.3);
        border-radius: 18px; user-select: none;
        -webkit-user-select: none; /* Safari */
        -ms-user-select: none; /* IE 10+ */
    }}
    .{pre}_btn:hover {{
        box-shadow: box-shadow: 0 3px 10px rgb(0,0,0,0.6);
        background: #4caf50; color: #fff;
    }}
</style>
<div>
    <div style="display: flex; flex-direction: row; padding: 8px">
        <div id="{pre}_prevBtn" class="{pre}_btn">Prev</div>
        <div id="{pre}_nextBtn" class="{pre}_btn">Next</div>
    </div>
    <div id="{pre}_status" style="padding: 10px"></div>
</div>
<div id="{pre}_imgContainer">
    ${{{dataIdx}.map((x, i) => "<div id='{pre}_content" + i + "'>" + x + "</div>").join("")}}
</div>
%3Cscript%3E
    (async () => {{
        const {pre}_n = ${{{dataIdx}.length}}; {pre}_imgIdx = 0;
        function {pre}_updatePageCount() {{
            if ({pre}_n > 0) document.querySelector("#{pre}_status").innerHTML = "Page: " + ({pre}_imgIdx + 1) + "/" + {pre}_n;
            else document.querySelector("#{pre}_status").innerHTML = "Page: 0/0"
        }}
        function {pre}_display() {{
            for (let i = 0; i < {pre}_n; i++) document.querySelector("#{pre}_content" + i).style.display = "none";
            if ({pre}_n > 0) document.querySelector("#{pre}_content" + {pre}_imgIdx).style.display = "block";
            {pre}_updatePageCount();
        }};
        document.querySelector("#{pre}_prevBtn").onclick = () => {{
            {pre}_imgIdx -= 1;
            {pre}_imgIdx = Math.max({pre}_imgIdx, 0);
            {pre}_display();
        }};
        document.querySelector("#{pre}_nextBtn").onclick = () => {{
            {pre}_imgIdx += 1;
            {pre}_imgIdx = Math.min({pre}_imgIdx, {pre}_n - 1);
            {pre}_display();
        }};
        {pre}_display();
    }})();
%3C/script%3E`) }}
<!-- k1lib.Carousel end -->""", fIdx                                             # Carousel
k1lib.cli.init.addAtomic(Carousel)                                               # Carousel
class Toggle(cli.BaseCli):                                                       # Toggle
    _idx = k1lib.AutoIncrement.random()                                          # Toggle
    def __init__(self):                                                          # Toggle
        """Button to toggle whether the content is displayed or
not. Useful if the html content is very big in size. Example::

    x = np.linspace(-2, 2); plt.plot(x, x ** 2)
    plt.gcf() | toImg() | toHtml() | viz.Toggle()

This will plot a graph, then create a button where you can toggle the image's visibility""" # Toggle
        self.content:str = "" # html string                                      # Toggle
        self._enteredRor = False                                                 # Toggle
    def __ror__(self, it): self._enteredRor = True; self.content = it if isinstance(it, str) else it | cli.toHtml(); return self # Toggle
    def __or__(self, it): # see discussion on Carousel()                         # Toggle
        if self._enteredRor: return it.__ror__(self)                             # Toggle
        else: return super().__or__(it)                                          # Toggle
    def _repr_html_(self):                                                       # Toggle
        pre = f"k1t_{Toggle._idx()}"                                             # Toggle
        html = f"""<!-- k1lib.Toggle start -->
<style>
    #{pre}_btn {{
        cursor: pointer;
        padding: 6px 12px;
        background: #eee;
        margin-right: 5px;
        color: #000;
        user-select: none;
        -webkit-user-select: none; /* Safari */
        -ms-user-select: none; /* IE 10+ */
        box-shadow: 0 3px 5px rgb(0,0,0,0.3);
        border-radius: 18px;
    }}
    #{pre}_btn:hover {{
        box-shadow: 0 3px 5px rgb(0,0,0,0.6);
        background: #4caf50;
        color: #fff;
    }}
</style>
<div>
    <div style="display: flex; flex-direction: row; padding: 4px">
        <div id="{pre}_btn">Show content</div>
        <div style="flex: 1"></div>
    </div>
    <div id="{pre}_content" style="display: none; margin-top: 12px">{self.content}</div>
</div>
<script>
    console.log("setup script ran for {pre}");
    {pre}_btn = document.querySelector("#{pre}_btn");
    {pre}_content = document.querySelector("#{pre}_content");

    {pre}_displayed = false;
    {pre}_btn.onclick = () => {{
        {pre}_displayed = !{pre}_displayed;
        {pre}_btn.innerHTML = {pre}_displayed ? "Hide content" : "Show content";
        {pre}_content.style.display = {pre}_displayed ? "block" : "none";
    }};
</script>
<!-- k1lib.Toggle end -->"""                                                     # Toggle
        return html                                                              # Toggle
    def _jsF(self, meta):                                                        # Toggle
        fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto(); pre = cli.init._jsDAuto() # Toggle
        return f"""const {fIdx} = ({dataIdx}) => {{
        return unescape(`
<!-- k1lib.Toggle start -->
<style>
    #{pre}_btn {{
        cursor: pointer;
        padding: 6px 12px;
        background: #eee;
        margin-right: 5px;
        color: #000;
        user-select: none;
        -webkit-user-select: none; /* Safari */
        -ms-user-select: none; /* IE 10+ */
        box-shadow: 0 3px 5px rgb(0,0,0,0.3);
        border-radius: 18px;
    }}
    #{pre}_btn:hover {{
        box-shadow: 0 3px 5px rgb(0,0,0,0.6);
        background: #4caf50;
        color: #fff;
    }}
</style>
<div>
    <div style="display: flex; flex-direction: row; padding: 4px">
        <div id="{pre}_btn">Show content</div>
        <div style="flex: 1"></div>
    </div>
    <div id="{pre}_content" style="display: none; margin-top: 12px">${{{dataIdx}}}</div>
</div>
%3Cscript%3E
    (async () => {{
        console.log("setup script ran for {pre}");
        {pre}_btn = document.querySelector("#{pre}_btn");
        {pre}_content = document.querySelector("#{pre}_content");

        {pre}_displayed = false;
        {pre}_btn.onclick = () => {{
            {pre}_displayed = !{pre}_displayed;
            {pre}_btn.innerHTML = {pre}_displayed ? "Hide content" : "Show content";
            {pre}_content.style.display = {pre}_displayed ? "block" : "none";
        }};
    }})();
%3C/script%3E`) }}
<!-- k1lib.Toggle end -->""", fIdx                                               # Toggle
k1lib.cli.init.addAtomic(Toggle)                                                 # Toggle
def ToggleImage():                                                               # ToggleImage
    """This function is sort of legacy. It's just ``img | toHtml() | viz.Toggle()`` really""" # ToggleImage
    return cli.toHtml() | Toggle()                                               # ToggleImage
class Scroll(cli.BaseCli):                                                       # Scroll
    _idx = k1lib.AutoIncrement.random()                                          # Scroll
    def __init__(self, height=300):                                              # Scroll
        """Creates a new preview html component. If content
is too long, then it will only show the first 500px, then
have a button to expand and view the rest. Example::

    x = np.linspace(-2, 2); plt.plot(x, x ** 2)
    plt.gcf() | toImg() | toHtml() | viz.Scroll()

This will plot a preview of a graph
:param height: height of the parent container"""                                 # Scroll
        self.content:str = "" # html string                                      # Scroll
        self.height = height; self._enteredRor = False                           # Scroll
    def __ror__(self, it): self._enteredRor = True; self.content = it; return self # Scroll
    def __or__(self, it): # see discussion on Carousel()                         # Scroll
        if self._enteredRor: return it.__ror__(self)                             # Scroll
        else: return super().__or__(it)                                          # Scroll
    def _repr_html_(self):                                                       # Scroll
        pre = f"k1scr_{Scroll._idx()}"                                           # Scroll
        html = f"""<!-- k1lib.Scroll start -->
<div style="max-height: {self.height}px; overflow-y: auto">{self.content}</div>
<!-- k1lib.Scroll end -->"""                                                     # Scroll
        return html                                                              # Scroll
    def _jsF(self, meta):                                                        # Scroll
        fIdx = cli.init._jsFAuto(); dataIdx = cli.init._jsDAuto(); pre = cli.init._jsDAuto() # Scroll
        return f"""const {fIdx} = ({dataIdx}) => {{
        return unescape(`<!-- k1lib.Scroll start -->
<div style="max-height: {self.height}px; overflow-y: auto">${{{dataIdx}}}</div>
<!-- k1lib.Scroll end -->`) }}""", fIdx                                          # Scroll
k1lib.cli.init.addAtomic(Scroll)                                                 # Scroll
def confusionMatrix(matrix:torch.Tensor, categories:List[str]=None, **kwargs):   # confusionMatrix
    """Plots a confusion matrix. Example::

    k1lib.viz.confusionMatrix(torch.rand(5, 5), ["a", "b", "c", "d", "e"])

.. image:: images/confusionMatrix.png

:param matrix: 2d matrix of shape (n, n)
:param categories: list of string categories
:param kwargs: keyword args passed into :meth:`plt.figure`"""                    # confusionMatrix
    if isinstance(matrix, torch.Tensor): matrix = matrix.numpy()                 # confusionMatrix
    if categories is None: categories = [f"{e}" for e in range(len(matrix))]     # confusionMatrix
    fig = plt.figure(**{"dpi":100, **kwargs}); ax = fig.add_subplot(111)         # confusionMatrix
    cax = ax.matshow(matrix); fig.colorbar(cax)                                  # confusionMatrix
                                                                                 # confusionMatrix
    with k1lib.ignoreWarnings():                                                 # confusionMatrix
        ax.set_xticklabels([''] + categories, rotation=90)                       # confusionMatrix
        ax.set_yticklabels([''] + categories)                                    # confusionMatrix
                                                                                 # confusionMatrix
    # Force label at every tick                                                  # confusionMatrix
    ax.xaxis.set_major_locator(mpl.ticker.MultipleLocator(1))                    # confusionMatrix
    ax.yaxis.set_major_locator(mpl.ticker.MultipleLocator(1))                    # confusionMatrix
    ax.xaxis.set_label_position('top')                                           # confusionMatrix
    plt.xlabel("Predictions"); plt.ylabel("Ground truth")                        # confusionMatrix
def FAnim(fig, f, frames, *args, **kwargs):                                      # FAnim
    """Matplotlib function animation, 60fps. Example::

    # line below so that the animation is displayed in the notebook. Included in :mod:`k1lib.imports` already, so you don't really have to do this!
    plt.rcParams["animation.html"] = "jshtml"
    x = np.linspace(-2, 2); y = x**2
    fig, ax = plt.subplots()
    plt.close() # close cause it'll display 1 animation, 1 static if we don't do this
    def f(frame):
        ax.clear()
        ax.set_ylim(0, 4); ax.set_xlim(-2, 2)
        ax.plot(x[:frame], y[:frame])
    k1lib.FAnim(fig, f, len(x)) # plays animation in cell

:param fig: figure object from `plt.figure(...)` command
:param f: function that accepts 1 frame from `frames`.
:param frames: number of frames, or iterator, to pass into function"""           # FAnim
    return partial(mpl.animation.FuncAnimation, interval=1000/30)(fig, f, frames, *args, **kwargs) # FAnim
from k1lib.cli import op                                                         # FAnim
def mask(img:torch.Tensor, act:torch.Tensor) -> torch.Tensor:                    # mask
    """Shows which part of the image the network is focusing on.

:param img: the image, expected to have dimension of (3, h, w)
:param act: the activation, expected to have dimension of (x, y), and with
    elements from 0 to 1."""                                                     # mask
    *_, h, w = img.shape                                                         # mask
    mask = act[None,] | nn.AdaptiveAvgPool2d([h//16, w//16]) | nn.AdaptiveAvgPool2d([h//8, w//8]) | nn.AdaptiveAvgPool2d([h, w]) # mask
    return mask * img | op().permute(1, 2, 0)                                    # mask
class PDF(object):                                                               # PDF
    def __init__(self, pdf:str=None, size=(700,500)):                            # PDF
        """Displays pdf in the notebook.
Example::

    viz.PDF("a.pdf")
    "a.pdf" | viz.PDF()

    viz.PDF("a.pdf", (700, 500))
    "a.pdf" | viz.PDF(size=(700, 500))

If you're exporting this notebook as html, then you have to make sure
you place the generated html file in the correct directory so that it
can reference those pdf files.

:param pdf: relative path to pdf file"""                                         # PDF
        self.pdf = pdf; self.size = size                                         # PDF
    def __ror__(self, pdf): self.pdf = pdf; return self                          # PDF
    def _repr_html_(self): return '<iframe src={0} width={1[0]} height={1[1]}></iframe>'.format(self.pdf, self.size) # PDF
    def _repr_latex_(self): return r'\includegraphics[width=1.0\textwidth]{{{0}}}'.format(self.pdf) # PDF



# AUTOGENERATED FILE! PLEASE DON'T EDIT HERE. EDIT THE SOURCE NOTEBOOKS INSTEAD
import k1lib, time, torch; from k1lib import fmt, cli
from k1lib.callbacks import Callback, Cbs
_ltime = 15; _lt1 = 8; _lt2 = 18
class TimeData:                                                                  # TimeData
    def __init__(self, tProfiler, mS:k1lib.selector.ModuleSelector):             # TimeData
        self.tProfiler = tProfiler; self.mS = mS                                 # TimeData
        self.startTime = None; self.time = 0; self.hook()                        # TimeData
    def hook(self):                                                              # TimeData
        def fpHk(m, i):                                                          # TimeData
            if self.tProfiler.is_cuda: torch.cuda.synchronize()                  # TimeData
            self.startTime = time.time()                                         # TimeData
        def fHk(m, i, o):                                                        # TimeData
            if self.tProfiler.is_cuda: torch.cuda.synchronize()                  # TimeData
            self.time += time.time() - self.startTime                            # TimeData
        self.fpH = self.mS.nn.register_forward_pre_hook(fpHk)                    # TimeData
        self.fH = self.mS.nn.register_forward_hook(fHk)                          # TimeData
    def unhook(self):                                                            # TimeData
        self.tProfiler.totalTime = max(self.tProfiler.totalTime, self.time)      # TimeData
        self.fpH.remove(); self.fH.remove()                                      # TimeData
    def __getstate__(self):                                                      # TimeData
        answer = dict(self.__dict__)                                             # TimeData
        del answer["mS"]; del answer["tProfiler"]; return answer                 # TimeData
    def __setstate__(self, state): self.__dict__.update(dict(state))             # TimeData
    def __str__(self):                                                           # TimeData
        if self.time <= 1e-20: return ""                                         # TimeData
        a = f"{fmt.time(self.time)}".ljust(_ltime)                               # TimeData
        b = f"{round(100 * self.time / self.tProfiler.totalTime)}%".rjust(_lt1)  # TimeData
        c = f"{round(100 * self.time / self.tProfiler.selectedMaxTime)}%".rjust(_lt2) if self.tProfiler.selectedMaxTime != None and "_timeProf_" in self.mS else "" # TimeData
        return f"{a}{b}{c}"                                                      # TimeData
class TimeProfiler(Callback):                                                    # TimeProfiler
    """Profiles execution time. Only measures forward times, as
backward times can't really be measured. Example::

    l = k1lib.Learner.sample()
    l.cbs.add(Cbs.Profiler())
    # views table
    l.Profiler.time
    # views table highlighted
    l.Profiler.time.css("#lin1")"""                                              # TimeProfiler
    def startRun(self):                                                          # TimeProfiler
        if not hasattr(self, "selector"): # if no selectors found                # TimeProfiler
            self.selector = self.l.model.select("")                              # TimeProfiler
        for m in self.selector.modules(): m.data = TimeData(self, m)             # TimeProfiler
        self.selector.displayF = lambda m: (fmt.txt.red if "_timeProf_" in m else fmt.txt.identity)(m.data) # TimeProfiler
        self.totalTime = 0; self.selectedMaxTime = None                          # TimeProfiler
    def startStep(self): return True                                             # TimeProfiler
    def _run(self):                                                              # TimeProfiler
        """Runs everything"""                                                    # TimeProfiler
        with self.cbs.context(), self.cbs.suspendEval():                         # TimeProfiler
            self.is_cuda = next(self.l.model.parameters()).is_cuda               # TimeProfiler
            if self.is_cuda: self.cbs.add(Cbs.Cuda())                            # TimeProfiler
            else: self.cbs.add(Cbs.Cpu())                                        # TimeProfiler
            self.l.run(1, 1)                                                     # TimeProfiler
        for m in self.selector.modules(): m.data.unhook()                        # TimeProfiler
    def css(self, css:str):                                                      # TimeProfiler
        """Selects a small part of the network to highlight. See also: :mod:`k1lib.selector`.""" # TimeProfiler
        self.selector.parse(k1lib.selector.preprocess(css, "_timeProf_"))        # TimeProfiler
        self.selectedMaxTime = 0                                                 # TimeProfiler
        for m in self.selector.modules():                                        # TimeProfiler
            if "_timeProf_" in m:                                                # TimeProfiler
                self.selectedMaxTime = max(self.selectedMaxTime, m.data.time)    # TimeProfiler
        print(self.__repr__())                                                   # TimeProfiler
        self.selector.clearProps(); self.selectedMaxTime = None                  # TimeProfiler
    def __repr__(self):                                                          # TimeProfiler
        header = "time".ljust(_ltime) + "% total".rjust(_lt1) + ("% selected max" if self.selectedMaxTime != None else "").rjust(_lt2) # TimeProfiler
        footer = ""                                                              # TimeProfiler
        if self.selectedMaxTime != None:                                         # TimeProfiler
            b = f"{round(100 * self.selectedMaxTime / self.totalTime)}%".rjust(_lt1, " ") # TimeProfiler
            st = f"{fmt.time(self.selectedMaxTime)}".rjust(_lt2)                 # TimeProfiler
            footer = ("Selected max", " " * _ltime + b + st)                     # TimeProfiler
        c = self.selector.__repr__(intro=False, header=header, footer=footer).split("\n") | cli.tab() | cli.join("\n") # TimeProfiler
        return f"""TimeProfiler ({"GPU" if self.is_cuda else "CPU"}):\n{c}

Caveats: This one's a bit stranger than memory and computation profilers
1. There is no "total" time (adding all times in all modules). There
    is total network execution time tho, which is just the time taken
    for the top level module to execute
2. "% selected max" column displays percentage of selected max, not
    percentage of total selected time, which may matter in your analysis

Can...
- tp.css("..."): highlights a particular part of the network
- tp.selector: to get internal k1lib.ModuleSelector object"""                    # TimeProfiler
